from __future__ import annotations

import functools
from dataclasses import dataclass

from garaga import garaga_rs
from garaga.algebra import Fp2, FunctionFelt, Polynomial, PyFelt, RationalFunction, T
from garaga.definitions import CURVES, CurveID, G1Point, G2Point, get_base_field
from garaga.hints.neg_3 import (
    construct_digit_vectors,
    neg_3_base_le,
    positive_negative_multiplicities,
)
from garaga.poseidon_transcript import hades_permutation


def get_field_type_from_ec_point(P: G1Point | G2Point) -> type[T]:
    """
    Maps an elliptic curve point to the type of the field it belongs to.
    G1Point -> PyFelt
    G2Point -> Fp2
    """
    if isinstance(P, G1Point):
        return PyFelt
    elif isinstance(P, G2Point):
        return Fp2
    else:
        raise ValueError(f"Invalid point type {type(P)}")


def get_ec_group_class_from_ec_point(P: G1Point | G2Point) -> type[G1Point | G2Point]:
    """
    Maps an elliptic curve point to the class of the elliptic curve group it belongs to.
    G1Point -> G1Point
    G2Point -> G2Point
    """
    if isinstance(P, G1Point):
        return G1Point
    elif isinstance(P, G2Point):
        return G2Point
    else:
        raise ValueError(f"Invalid point type {type(P)}")


def derive_ec_point_from_X(
    x: PyFelt | int | Fp2, curve_id: CurveID
) -> tuple[PyFelt, PyFelt, list[PyFelt]] | tuple[Fp2, Fp2, list[Fp2]]:
    """
    From a "random" x coordinate (in practice obtained via the Cairo Poseidon252 hash), finds via a
    "try-and-increment" algorithm a point on the curve for a given curveID.
    Works for curves over base field ("PyFelt") or degree 2 extension field ("Fp2")
    Returns :
    - x (PyFelt or Fp2) - x coordinate of the obtained point after the try-and-increment
    - y (PyFelt or Fp2) - y coordinate of the obtained point after the try-and-increment
    - g_rhs_roots (list of PyFelt of Fp2). A list of square roots over the given field.

    At each attempt, if rhs(x) = x^3 + ax + b is not a quad residue (ie: the point is not on the curve),
    x is updated by hashing it with the attempt : new_x = poseidon(x, attempt)

    Since in a finite field, if z is not a quad residue, g*z is a quad residue (alternatively), at each attempt,
    the square root of g*x is stored in the g*rhs_roots array.

    This is used to verify the existence of the square roots in Cairo.
    See the derive_ec_point_from_X cairo function in ec_ops.cairo.
    """
    field = get_base_field(curve_id.value)
    if isinstance(x, int):
        x = field(x)

    def rhs_compute(x: PyFelt | Fp2) -> PyFelt | Fp2:
        """
        Compute the right hand side of the Weirstrass equation.
        rhs(x) = x^3 + ax + b
        """
        if isinstance(x, Fp2):
            return (
                x**3
                + x * field(CURVES[curve_id.value].a)
                + Fp2(
                    field(CURVES[curve_id.value].b20), field(CURVES[curve_id.value].b21)
                )
            )
        else:
            return (
                x**3
                + field(CURVES[curve_id.value].a) * x
                + field(CURVES[curve_id.value].b)
            )

    if isinstance(x, Fp2):
        g = Fp2(
            field(CURVES[curve_id.value].nr_a0),
            field(CURVES[curve_id.value].nr_a1),
        )
    else:
        g = field(CURVES[curve_id.value].fp_generator)

    rhs = rhs_compute(x)
    g_rhs_roots = []
    attempt = 0
    while not rhs.is_quad_residue():
        g_rhs = rhs * g
        g_rhs_roots.append(g_rhs.sqrt())
        if isinstance(x, Fp2):
            _s0, _, _ = hades_permutation(x.a0.value, x.a1.value, 2)
            _x0, _x1, _ = hades_permutation(_s0, attempt, 2)
            x = Fp2(field(_x0), field(_x1))
        else:
            _x, _, _ = hades_permutation(x.value, attempt, 2)
            x = field(_x)

        rhs = rhs_compute(x)
        attempt += 1

    y = rhs.sqrt()
    assert y**2 == rhs
    return x, y, g_rhs_roots


def zk_ecip_hint(
    Bs: list[G1Point] | list[G2Point], scalars: list[int], use_rust: bool = True
) -> tuple[G1Point | G2Point, FunctionFelt[T]]:
    """
    Inputs:
    - Bs: list of points on the curve
    - scalars: list of scalars
    Returns:
    - Q: MSM of Bs by scalars contained in dss matrix
    - sum_dlog: sum of the logarithmic derivatives of the functions in Ds

    Use this to verify equation 3 in https://eprint.iacr.org/2022/596.pdf
    Partial Ref : https://gist.github.com/Liam-Eagen/666d0771f4968adccd6087465b8c5bd4
    Full algo verifying it in the function verify_ecip below.
    """
    assert len(Bs) == len(scalars)

    ec_group_class = get_ec_group_class_from_ec_point(Bs[0])
    if ec_group_class == G1Point and use_rust:
        pts = []
        c_id = Bs[0].curve_id
        for pt in Bs:
            pts.extend([pt.x, pt.y])
        field_type = get_field_type_from_ec_point(Bs[0])
        field = get_base_field(c_id.value, field_type)

        q, a_num, a_den, b_num, b_den = garaga_rs.zk_ecip_hint(
            pts, list(scalars), c_id.value
        )

        a_num = [field(f) for f in a_num] if len(a_num) > 0 else [field.zero()]
        a_den = [field(f) for f in a_den] if len(a_den) > 0 else [field.one()]
        b_num = [field(f) for f in b_num] if len(b_num) > 0 else [field.zero()]
        b_den = [field(f) for f in b_den] if len(b_den) > 0 else [field.one()]

        Q = G1Point(q[0], q[1], c_id)
        sum_dlog = FunctionFelt(
            RationalFunction(Polynomial(a_num), Polynomial(a_den)),
            RationalFunction(Polynomial(b_num), Polynomial(b_den)),
        )
    else:
        dss = construct_digit_vectors(scalars)
        Q, Ds = ecip_functions(Bs, dss)
        dlogs = [dlog(D) for D in Ds]
        sum_dlog = dlogs[0]
        field = get_base_field(Q.curve_id.value, PyFelt)
        for i in range(1, len(dlogs)):
            sum_dlog = sum_dlog + (-3) ** i * dlogs[i]
    return Q, sum_dlog


def verify_ecip(
    Bs: list[G1Point] | list[G2Point],
    scalars: list[int],
    Q: G1Point | G2Point = None,
    sum_dlog: FunctionFelt[T] = None,
    A0: G1Point | G2Point = None,
    use_rust: bool = True,
) -> bool:
    """
    Verifies the zk-ecip hint.
    If Q, sum_dlog are not provided from a previous computation of the zk_ecip_hint, it will compute them.
    If the random point A0 is not provided for verifying the hint, a random one will be sampled.
    """
    # Prover :
    if Q is None or sum_dlog is None:
        Q, sum_dlog = zk_ecip_hint(Bs, scalars, use_rust)
    else:
        Q = Q
        sum_dlog = sum_dlog
    # Verifier :
    assert sum_dlog.validate_degrees(len(Bs))
    epns = [
        positive_negative_multiplicities(neg_3_base_le(scalar)) for scalar in scalars
    ]

    c_id = Q.curve_id.value
    ec_group_class = get_ec_group_class_from_ec_point(Q)
    field_type = get_field_type_from_ec_point(Q)

    field = get_base_field(Q.curve_id.value, field_type)

    # Verifier must derive a random point A0.
    # In non-interactive context, hash public inputs (Bs, scalars) and prover info (Q, sum_dlog) then map the resulting field element to a point.
    if A0 is None:
        A0 = ec_group_class.gen_random_point(Q.curve_id)
    else:
        A0 = A0

    A2 = A0.scalar_mul(-2)

    xA0 = field(A0.x)
    yA0 = field(A0.y)

    xA2 = field(A2.x)
    yA2 = field(A2.y)

    mA0, bA0 = slope_intercept(A0, A0)
    mA0A2, _ = slope_intercept(A2, A0)

    coeff2 = (2 * yA2 * (xA0 - xA2)) / (
        3 * xA2**2 + field(CURVES[c_id].a) - 2 * mA0A2 * yA2
    )
    coeff0 = coeff2 + 2 * mA0A2

    basis_sum = field.zero()
    for i, (P, (ep, en)) in enumerate(zip(Bs, epns)):
        if P.is_infinity():
            # print("skipping infinity")
            continue
        if ep - en == 0:
            # print("skipping 0")
            continue
        basis_sum += eval_point_challenge_signed(P, xA0, mA0, bA0, ep, en)
        # print(f"rhs_acc {i}: {basis_sum.value}")

    if Q.is_infinity():
        RHS = basis_sum
    else:
        RHS = eval_point_challenge(-Q, xA0, mA0, bA0, 1) + basis_sum

    LHS = coeff0 * sum_dlog.evaluate(xA0, yA0) - coeff2 * sum_dlog.evaluate(xA2, yA2)

    # Verifier must check that LHS = RHS.
    assert LHS == RHS, f"LHS: {LHS}, RHS: {RHS}"

    #########
    assert Q == ec_group_class.msm(
        Bs, scalars
    )  # Sanity check. Not part of the verification.
    ##########
    return True


def slope_intercept(
    P: G1Point | G2Point, Q: G1Point | G2Point
) -> tuple[PyFelt, PyFelt] | tuple[Fp2, Fp2]:
    """
    Returns the slope and intercept of the line passing through points P and Q.
    y = mx + b
    """
    field = get_base_field(P.curve_id.value, get_field_type_from_ec_point(P))
    if P == Q:
        px, py = field(P.x), field(P.y)
        m = (3 * px**2 + field(CURVES[P.curve_id.value].a)) / (2 * py)
        b = py - m * px
        return (m, b)
    else:
        px, py = field(P.x), field(P.y)
        qx, qy = field(Q.x), field(Q.y)
        m = (py - qy) / (px - qx)
        b = qy - m * qx
        return (m, b)


def eval_point_challenge(
    P: G1Point | G2Point, xA0, mA0, bA0, multiplicity: int
) -> PyFelt | Fp2:
    field_type = get_field_type_from_ec_point(P)
    field = get_base_field(P.curve_id.value, field_type)

    xP, yP = field(P.x), field(P.y)
    num = xA0 - xP
    den = yP - mA0 * xP - bA0
    res = multiplicity * num / den
    assert isinstance(res, field_type), f"Expected {field_type}, got {type(res)}"
    return res


def eval_point_challenge_signed(
    P: G1Point | G2Point, xA0, mA0, bA0, ep: int, en: int
) -> PyFelt | Fp2:
    return eval_point_challenge(P, xA0, mA0, bA0, ep) + eval_point_challenge(
        -P, xA0, mA0, bA0, en
    )


def line(P: G1Point | G2Point, Q: G1Point | G2Point) -> FF[T]:
    """
    Returns line function passing through points, works for all points and returns 1 for O + O = O
    """
    assert (
        P.curve_id == Q.curve_id
    ), f"Points must be on the same curve, got {P.curve_id} and {Q.curve_id}"
    assert isinstance(
        P, type(Q)
    ), f"Points must be in the same group, got {type(P)} and {type(Q)}"

    if isinstance(P, G1Point):
        field_type = PyFelt
    elif isinstance(P, G2Point):
        field_type = Fp2
    else:
        raise ValueError(f"Invalid point type {type(P)}")

    field = get_base_field(P.curve_id.value, field_type)

    if P.is_infinity():
        if Q.is_infinity():
            return FF([Polynomial([field.one()])], P.curve_id)
        else:
            Qx = field(Q.x)
            return FF([Polynomial([-Qx, field.one()])], P.curve_id)
    if Q.is_infinity():
        Px = field(P.x)
        return FF([Polynomial([-Px, field.one()])], P.curve_id)

    Px, Py = field(P.x), field(P.y)

    if P == Q:
        m = (3 * Px**2 + field(CURVES[P.curve_id.value].a)) / (2 * Py)
        b = Py - m * Px
        # -m*x + y -b
        return FF([Polynomial([-b, -m]), Polynomial([field.one()])], P.curve_id)

    if P == -Q:
        return FF([Polynomial([-Px, field.one()])], P.curve_id)

    Qx, Qy = field(Q.x), field(Q.y)

    m = (Py - Qy) / (Px - Qx)
    b = Qy - m * Qx
    return FF([Polynomial([-b, -m]), Polynomial([field.one()])], P.curve_id)


@dataclass
class FF:
    """
    Represents a polynomial over F_p[x] or F_p^2[x]
    Example : F(x, y) = c0(x) + c1(x) * y + c2(x) * y^2 + ...
    where c0, c1, c2, ... are polynomials over F_p[x] or F_p^2[x]
    Used to represent a subset of the Function Field where coefficients are polynomials instead of rational functions.
    """

    coeffs: list[Polynomial[T]]
    y2: Polynomial[T]  # = x^3 + ax + b, where a, b are the curve parameters
    p: int
    curve_id: CurveID
    type: type[T]  # PyFelt or Fp2

    def __init__(self, coeffs: list[Polynomial[T]], curve_id: CurveID):
        self.coeffs = coeffs
        self.p = coeffs[0][0].p
        self.field = get_base_field(curve_id.value, type(coeffs[0][0]))
        self.curve_id = curve_id
        self.type = type(coeffs[0][0])
        a = self.field(CURVES[curve_id.value].a)
        if self.type == PyFelt:
            b = self.field(CURVES[curve_id.value].b)
        else:
            b = self.field((CURVES[curve_id.value].b20, CURVES[curve_id.value].b21))

        # y² = x³ + ax + b
        self.y2 = Polynomial([b, a, self.field.zero(), self.field.one()])

    def degree(self) -> int:
        return len(self.coeffs) - 1

    def __getitem__(self, i: int) -> Polynomial:
        try:
            return self.coeffs[i]
        except IndexError:
            return Polynomial.zero(self.p, self.type)

    def __add__(self, other: FF) -> FF:
        if not isinstance(other, FF):
            raise TypeError(f"Cannot add FF and {type(other)}")

        ns, no = len(self.coeffs), len(other.coeffs)
        if ns >= no:
            coeffs = self.coeffs[:]
            for i in range(no):
                coeffs[i] = coeffs[i] + other.coeffs[i]
        else:
            coeffs = other.coeffs[:]
            for i in range(ns):
                coeffs[i] = coeffs[i] + self.coeffs[i]

        return FF(coeffs, self.curve_id)

    def __mul__(self, other: "FF" | Polynomial | PyFelt) -> "FF":
        if isinstance(other, (PyFelt, Polynomial)):
            return FF([c * other for c in self.coeffs], self.curve_id)
        elif not isinstance(other, FF):
            raise TypeError(
                f"Cannot multiply polynomial by type {type(other)}, must be PyFelt, Polynomial, or FF"
            )

        if self.coeffs == [] or other.coeffs == []:
            return FF([Polynomial([self.field.zero()])], self.curve_id)

        zero = Polynomial.zero(self.p, self.type)

        buf = [zero] * (len(self.coeffs) + len(other.coeffs) - 1)
        for i in range(len(self.coeffs)):
            if self.coeffs[i].is_zero():
                continue  # optimization for sparse polynomials
            for j in range(len(other.coeffs)):
                buf[i + j] = buf[i + j] + self.coeffs[i] * other.coeffs[j]
        res = FF(buf, self.curve_id)
        return res

    def neg_y(self) -> "FF":
        """Keeps all the coefficients the same but negate the degree 1 coeff"""
        if len(self.coeffs) < 2:
            return self
        coeffs = self.coeffs[:]
        coeffs[1] = -coeffs[1]
        return FF(coeffs, self.curve_id)

    def reduce(self) -> "FF":
        """
        Reduces the polynomial modulo y² = x³ + ax + b
        Replaces y² with x³ + ax + b and reduces the degree of the polynomial
        Always returns a polynomial of at most degree 1.
        Enforce it has 2 coefficients (padding with zero polynomials if needed)
        """
        if len(self.coeffs) <= 2:
            while len(self.coeffs) < 2:
                self.coeffs.append(Polynomial.zero(self.p, self.type))
            return self
        y2 = self.y2
        deg_0_coeff = self.coeffs[0]
        deg_1_coeff = self.coeffs[1]
        for i, poly in enumerate(self.coeffs):
            if i == 0 or i == 1:
                continue
            if i % 2 == 0:
                deg_0_coeff = deg_0_coeff + poly * y2
            else:
                deg_1_coeff = deg_1_coeff + poly * y2
                y2 = y2 * y2
        return FF([deg_0_coeff, deg_1_coeff], self.curve_id)

    def to_poly(self) -> Polynomial[T]:
        """
        "Downcasts" the FF to a Polynomial iff the FF is of degree 1.
        """
        assert len(self.coeffs) == 1
        return self.coeffs[0]

    def div_by_poly(self, poly: Polynomial[T]) -> "FF":
        return FF([c // poly for c in self.coeffs], self.curve_id)

    def normalize(self) -> "FF":
        """
        Normalize the polynomial by dividing all coefficients by the first one
        """
        coeff = self.coeffs[0][0]
        return FF([c * coeff.__inv__() for c in self.coeffs], self.curve_id)


class EmptyListOfPoints(Exception):
    pass


def construct_function(Ps: list[G1Point] | list[G2Point]) -> FF:
    """
    Returns a function field element (class FF) exactly interpolating the points Ps.
    """
    if len(Ps) == 0:
        raise EmptyListOfPoints(
            "Cannot construct a function from an empty list of points"
        )
    xs = [(P, line(P, -P)) for P in Ps]

    while len(xs) != 1:
        xs2 = []

        if len(xs) & 1:
            x0 = xs[0]
            xs = xs[1:]
        else:
            x0 = None

        for n in range(0, len(xs) // 2):
            (A, aNum) = xs[2 * n]
            (B, bNum) = xs[2 * n + 1]
            aNum_bNum = aNum * bNum
            line_AB = line(A, B)
            product = aNum_bNum * line_AB
            num = product.reduce()
            den = (line(A, -A) * line(B, -B)).to_poly()
            D = num.div_by_poly(den)
            xs2.append((A.add(B), D))

        if x0 is not None:
            xs2.append(x0)

        xs = xs2

    assert xs[0][0].is_infinity()

    return xs[-1][1].normalize()


def row_function(
    ds: list[int], Ps: list[G1Point] | list[G2Point], Q: G1Point | G2Point
) -> tuple[FF, G1Point | G2Point]:
    ec_group_class = G1Point if isinstance(Q, G1Point) else G2Point
    infinity = ec_group_class.infinity(Q.curve_id)

    digits_points = [
        P if d == 1 else -P if d == -1 else infinity for d, P in zip(ds, Ps)
    ]
    sum_digits_points = functools.reduce(lambda x, y: x.add(y), digits_points)
    Q2 = Q.scalar_mul(-3).add(sum_digits_points)
    Q_neg = -Q
    div_ = [Q_neg, Q_neg, Q_neg, -Q2] + digits_points
    div = [P for P in div_ if not P.is_infinity()]
    try:
        D = construct_function(div)
    except EmptyListOfPoints:
        field_type = get_field_type_from_ec_point(div_[0])
        # Returning 1 so that dLog(D) = 0 and doesn't contribute to the sum.
        D = FF(
            [Polynomial.one(get_base_field(Q.curve_id, field_type).p, field_type)],
            Q.curve_id,
        )

    return (D, Q2)


def ecip_functions(
    Bs: list[G1Point] | list[G2Point], dss: list[list[int]]
) -> tuple[G1Point | G2Point, list[FF]]:
    dss.reverse()
    ec_group_class = G1Point if isinstance(Bs[0], G1Point) else G2Point
    Q = ec_group_class.infinity(Bs[0].curve_id)
    Ds = []
    for i, ds in enumerate(dss):
        D, Q = row_function(ds, Bs, Q)
        Ds.append(D)

    Ds.reverse()
    return (Q, Ds)


def dlog(d: FF) -> FunctionFelt:
    """
    Compute the logarithmic derivative of a FunctionFieldElement (class FF),
    Returns a FunctionFelt such that F = a(x) + b(x) * y with a, b RationalFunctions

    Ref https://gist.github.com/Liam-Eagen/666d0771f4968adccd6087465b8c5bd4 (cell #12)

    Example values:
    D: (3*x + 11)*y + x^2 + x + 1
    Dx: 3*y + 2*x + 1
    Dy: 3*x + 11
    Dz_tmp: (9*x^3 + 33*x^2)/(2*y)
    Dz: (6*y^2 + (4*x + 2)*y + 9*x^3 + 33*x^2)/(2*y)
    U=2*y*Dz: 6*y^2 + (4*x + 2)*y + 9*x^3 + 33*x^2
    V=2*y*D: (6*x + 22)*y^2 + (2*x^2 + 2*x + 2)*y

    """
    field = d.field

    d: FF = d.reduce()
    assert len(d.coeffs) == 2, f"D has {len(d.coeffs)} coeffs: {d.coeffs}"

    Dx = FF([d[0].differentiate(), d[1].differentiate()], d.curve_id)
    Dy: Polynomial = d[1]  # B(x)

    TWO_Y: FF = FF(
        [Polynomial.zero(field.p, d.type), Polynomial([field(2)])], d.curve_id
    )
    U: FF = Dx * TWO_Y + FF(
        [
            Dy
            * Polynomial(
                [
                    field(CURVES[d.curve_id.value].a),
                    field.zero(),
                    field(3),
                ]  # 3x^2 + A
            ),
            Polynomial.zero(field.p, d.type),
        ],
        d.curve_id,
    )

    V: FF = TWO_Y * d

    # print(f"U: {print_ff(U)}")
    # print(f"V: {print_ff(V)}")

    Num: FF = (U * V.neg_y()).reduce()
    Den_FF: FF = (V * V.neg_y()).reduce()

    # print(f"Num: {print_ff(Num)}")
    # print(f"Den: {print_ff(Den_FF)}")

    assert Den_FF[
        1
    ].is_zero(), f"Den[1] is not zero: {Den_FF[1].print_as_sage_poly('x')}"

    Den: Polynomial = Den_FF[0]

    _, _, gcd_0 = Polynomial.xgcd(Num[0], Den)
    # print(f"GCD_0: {gcd_0.print_as_sage_poly('x')}")
    _, _, gcd_1 = Polynomial.xgcd(Num[1], Den)
    # print(f"GCD_1: {gcd_1.print_as_sage_poly('x')}")

    # Simplify the numerator and denominator by dividing by the gcd

    a_num = Num[0] // gcd_0
    a_den = Den // gcd_0
    b_num = Num[1] // gcd_1
    b_den = Den // gcd_1

    # Normalize to obtain exactly Sage's coeffs.
    res = FunctionFelt(
        a=RationalFunction(
            a_num * Den.leading_coefficient().__inv__(),
            a_den * a_den.leading_coefficient().__inv__(),
        ),
        b=RationalFunction(
            b_num * Den.leading_coefficient().__inv__(),
            b_den * b_den.leading_coefficient().__inv__(),
        ),
    )
    # print(f"res: {res.print_as_sage_poly()}\n")
    return res


def print_ff(ff: FF):
    string = ""
    coeffs = ff.coeffs
    for i, p in enumerate(coeffs[::-1]):
        coeff_str = p.print_as_sage_poly(var_name="x", as_hex=True)

        if i == len(coeffs) - 1:
            if coeff_str == "":
                string = string[:-2]
            else:
                string += f"{coeff_str}"
        elif i == len(coeffs) - 2:
            string += f"({coeff_str})*y + "
        else:
            string += f"({coeff_str})*y^{len(coeffs) - i - 1} + "
    # print(string)
    return string


def n_points_from_n_coeffs(n_coeffs: int, batched: bool) -> int:
    if batched:
        extra = 4 * 2
    else:
        extra = 0

    # n_coeffs = 10 + 4n_points => 4n_points = n_coeffs - 10
    assert n_coeffs >= 10 + extra
    assert (n_coeffs - 10 - extra) % 4 == 0
    return (n_coeffs - 10 - extra) // 4


def n_coeffs_from_n_points(n_points: int, batched: bool) -> tuple[int, int, int, int]:
    if batched:
        extra = 2
    else:
        extra = 0

    return (
        1 + n_points + extra,
        1 + n_points + 1 + extra,
        1 + n_points + 1 + extra,
        1 + n_points + 4 + extra,
    )


if __name__ == "__main__":
    import random

    from garaga.definitions import STARK
    from garaga.hints.io import int_array_to_u384_array, int_to_u384

    random.seed(0)

    def build_cairo1_tests_derive_ec_point_from_X(x: int, curve_id: CurveID, idx: int):
        x_f, y, roots = derive_ec_point_from_X(x, curve_id)

        code = f"""
        #[test]
            fn derive_ec_point_from_X_{CurveID(curve_id).name}_{idx}() {{
                let x: felt252 = {x%STARK};
                let y: u384 = {int_to_u384(y)};
                let grhs_roots:Array<u384> = {int_array_to_u384_array(roots)};
                let result = derive_ec_point_from_X(x, y, grhs_roots, {curve_id.value});
                assert!(result.x == {int_to_u384(x_f)});
                assert!(result.y == y);
            }}
            """
        return code

    # codes = "\n".join(
    #     [
    #         build_cairo1_tests_derive_ec_point_from_X(x, curve_id, idx)
    #         for idx, x in enumerate([random.randint(0, STARK - 1) for _ in range(2)])
    #         for curve_id in CurveID
    #     ]
    # )

    # print(codes)

    # average_n_roots = 0
    # max_n_roots = 0
    # n = 10000
    # for i in range(n):
    #     x, y, roots = derive_ec_point_from_X(
    #         Fp2.random(CURVES[0].p, max_value=STARK), CurveID(0)
    #     )
    #     # print(f"x: {x}, y: {y}, roots: {roots}")
    #     max_n_roots = max(max_n_roots, len(roots))
    #     average_n_roots += len(roots)
    # print(f"Average number of roots: {average_n_roots / n}")
    # print(f"Max number of roots: {max_n_roots}")

    # verify_ecip([G1Point.gen_random_point(CurveID.SECP256K1)], scalars=[-1])

    import time

    order = CURVES[CurveID.BN254.value].n
    n = 50
    Bs = [G1Point.gen_random_point(CurveID.BN254) for _ in range(n)]
    ss = [random.randint(1, order) for _ in range(n)]

    t0 = time.time()
    ZZ = zk_ecip_hint(Bs, ss, use_rust=False)
    time_taken_py = time.time() - t0
    print(f"Time taken py : {time_taken_py}")

    t0 = time.time()
    ZZ_rs = zk_ecip_hint(Bs, ss, use_rust=True)
    time_taken_rs = time.time() - t0
    print(f"Time taken rs : {time_taken_rs}")

    print(f"Ratio: {time_taken_py / time_taken_rs}")

    assert ZZ == ZZ_rs



"""
Tower based arithmetic for BN254 and BLS12-381 on Fq2, Fq6, Fq12.
"""

import random
from dataclasses import dataclass

from garaga import garaga_rs
from garaga.algebra import ModuloCircuitElement, Polynomial, PyFelt
from garaga.definitions import CURVES, direct_to_tower, get_base_field, tower_to_direct


@dataclass(slots=True)
class E2:
    a0: int
    a1: int
    p: int

    def __str__(self) -> str:
        return f"({(self.a0)} + {self.a1}*u)"

    def __eq__(self, other):
        return self.a0 == other.a0 and self.a1 == other.a1

    @staticmethod
    def zero(p: int):
        return E2(0, 0, p)

    @staticmethod
    def one(p: int):
        return E2(1, 0, p)

    @staticmethod
    def random(p: int):
        return E2(random.randint(0, p - 1), random.randint(0, p - 1), p)

    @property
    def felt_coeffs(self) -> list[PyFelt]:
        return [PyFelt(self.a0, self.p), PyFelt(self.a1, self.p)]

    def __add__(self, other):
        a0 = (self.a0 + other.a0) % self.p
        a1 = (self.a1 + other.a1) % self.p
        return E2(a0, a1, self.p)

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        if isinstance(other, E2):
            a0 = (self.a0 - other.a0) % self.p
            a1 = (self.a1 - other.a1) % self.p
            return E2(a0, a1, self.p)
        elif isinstance(other, int):
            return E2((self.a0 - other) % self.p, (self.a1 - other) % self.p, self.p)
        return NotImplemented

    def __rsub__(self, other):
        return self.__neg__().__add__(other)

    def __mul__(self, other):
        if isinstance(other, E2):
            a = (self.a0 + self.a1) * (other.a0 + other.a1) % self.p
            b, c = self.a0 * other.a0 % self.p, self.a1 * other.a1 % self.p
            return E2((b - c) % self.p, (a - b - c) % self.p, self.p)
        elif isinstance(other, int):
            return E2(self.a0 * other % self.p, self.a1 * other % self.p, self.p)
        return NotImplemented

    def __rmul__(self, other):
        return self.__mul__(other)

    def __inv__(self):
        t0, t1 = (self.a0 * self.a0 % self.p, self.a1 * self.a1 % self.p)
        t0 = (t0 + t1) % self.p
        t1 = pow(t0, -1, self.p)
        return E2(self.a0 * t1 % self.p, -(self.a1 * t1) % self.p, self.p)

    def div(self, other):
        if isinstance(other, E2):
            return self * other.__inv__()
        return NotImplemented

    def __pow__(self, p: int):
        """
        Compute x**p in F_p^2 using square-and-multiply algorithm.
        Args:
        p: The exponent, a non-negative integer.
        Returns:
        x**p in F_p^2, represented similarly as x.
        """
        assert isinstance(p, int) and p >= 0

        # Handle the easy cases.
        if p == 0:
            # x**0 = 1, where 1 is the multiplicative identity in F_p^2.
            return E2(1, 0, self.p)
        elif p == 1:
            # x**1 = x.
            return self

        # Start the computation.
        result = self.one(
            self.p
        )  # Initialize result as the multiplicative identity in F_p^2.
        temp = self  # Initialize temp as self.

        # Loop through each bit of the exponent p.
        for bit in reversed(bin(p)[2:]):  # [2:] to strip the "0b" prefix.
            if bit == "1":
                result = result * temp
            temp = temp * temp

        return result

    def __truediv__(self, other):
        if isinstance(other, E2):
            return self * other.__inv__()
        elif isinstance(other, int):
            return self * pow(other, -1, self.p)

        return NotImplemented

    def __rtruediv__(self, other):
        if isinstance(other, E2):
            return other * self.__inv__()
        elif isinstance(other, int):
            return other * self.__inv__()

        return NotImplemented

    def __neg__(self):
        return E2(-self.a0 % self.p, -self.a1 % self.p, self.p)

    def conjugate(self):
        return E2(self.a0, -self.a1 % self.p, self.p)


@dataclass(slots=True, init=False)
class E6:
    b0: E2
    b1: E2
    b2: E2
    curve_id: int
    non_residue: E2

    def __init__(self, x: list[int | PyFelt | E2], curve_id: int):
        curve = CURVES[curve_id]
        self.curve_id = curve_id
        self.non_residue = E2(curve.nr_a0, curve.nr_a1, curve.p)
        if isinstance(x[0], int):
            self.b0 = E2(x[0], x[1], curve.p)
            self.b1 = E2(x[2], x[3], curve.p)
            self.b2 = E2(x[4], x[5], curve.p)
        elif isinstance(x[0], E2):
            self.b0 = x[0]
            self.b1 = x[1]
            self.b2 = x[2]
        elif isinstance(x[0], PyFelt):
            self.b0 = E2(x[0].value, x[1].value, curve.p)
            self.b1 = E2(x[2].value, x[3].value, curve.p)
            self.b2 = E2(x[4].value, x[5].value, curve.p)
        else:
            raise ValueError(f"Invalid type {type(x[0])}")

    @property
    def coeffs(self) -> list[int]:
        return [self.b0.a0, self.b0.a1, self.b1.a0, self.b1.a1, self.b2.a0, self.b2.a1]

    @property
    def felt_coeffs(self) -> list[PyFelt]:
        return [PyFelt(c, self.b0.p) for c in self.coeffs]

    @property
    def value_coeffs(self) -> list[int]:
        return [
            self.b0.a0,
            self.b0.a1,
            self.b1.a0,
            self.b1.a1,
            self.b2.a0,
            self.b2.a1,
        ]

    @staticmethod
    def from_poly(poly: Polynomial, curve_id: int):
        field = get_base_field(curve_id)
        coeffs = poly.get_coeffs()
        coeffs = coeffs + [field(0)] * (6 - len(coeffs))
        coeffs = direct_to_tower(coeffs, curve_id, 6)
        return E6(coeffs, curve_id)

    def to_poly(self) -> Polynomial:
        field = get_base_field(self.curve_id)
        coeffs = [field(c) for c in self.value_coeffs]
        coeffs = tower_to_direct(coeffs, self.curve_id, 6)
        return Polynomial(coeffs)

    @staticmethod
    def zero(curve_id: int):
        p = CURVES[curve_id].p
        return E6([E2.zero(p), E2.zero(p), E2.zero(p)], curve_id)

    @staticmethod
    def one(curve_id: int):
        p = CURVES[curve_id].p
        return E6([E2.one(p), E2.zero(p), E2.zero(p)], curve_id)

    @staticmethod
    def random(curve_id: int):
        p = CURVES[curve_id].p
        return E6([E2.random(p), E2.random(p), E2.random(p)], curve_id)

    def __str__(self) -> str:
        return f"{self.b0} + {self.b1}*v + {self.b2}*v^2"

    def __add__(self, other):
        if isinstance(other, E6):
            return E6(
                [self.b0 + other.b0, self.b1 + other.b1, self.b2 + other.b2],
                self.curve_id,
            )
        raise NotImplementedError

    def __sub__(self, other):
        if isinstance(other, E6):
            return E6(
                [self.b0 - other.b0, self.b1 - other.b1, self.b2 - other.b2],
                self.curve_id,
            )
        raise NotImplementedError

    def __neg__(self):
        return E6([-self.b0, -self.b1, -self.b2], self.curve_id)

    def __mul__(self, other):
        x0, x1, x2 = self.b0, self.b1, self.b2
        y0, y1, y2 = other.b0, other.b1, other.b2

        t0 = x0 * y0
        t1 = x1 * y1
        t2 = x2 * y2

        c0 = t0 + self.non_residue * ((x1 + x2) * (y1 + y2) - t1 - t2)
        c1 = (x0 + x1) * (y0 + y1) - t0 - t1 + self.non_residue * t2
        c2 = (x0 + x2) * (y0 + y2) - t0 - t2 + t1
        return E6([c0, c1, c2], self.curve_id)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __inv__(self):
        t0, t1, t2 = self.b0 * self.b0, self.b1 * self.b1, self.b2 * self.b2
        t3, t4, t5 = self.b0 * self.b1, self.b0 * self.b2, self.b1 * self.b2
        c0 = t0 - self.non_residue * t5
        c1 = self.non_residue * t2 - t3
        c2 = t1 - t4
        t6 = self.b0 * c0
        d1 = self.b2 * c1
        d2 = self.b1 * c2
        d1 = self.non_residue * (d1 + d2)
        t6 = t6 + d1
        t6 = t6.__inv__()
        return E6([c0 * t6, c1 * t6, c2 * t6], self.curve_id)

    def mul_by_non_residue(self):
        return E6([self.non_residue * self.b2, self.b0, self.b1], self.curve_id)

    def square_torus(self):
        # SQ = 1/2(x+ v/x) <=> v = x (2SQ - x)
        one_over_2 = E6([pow(2, -1, self.b0.p), 0, 0, 0, 0, 0], self.curve_id)
        v = E6([0, 0, 1, 0, 0, 0], self.curve_id)
        x_inv = self.__inv__()
        v_x_inv = v * x_inv
        sq = one_over_2 * (self + v_x_inv)
        return sq

    def div(self, other):
        if isinstance(other, E6):
            return self * other.__inv__()
        return NotImplementedError

    def __pow__(self, p: int):
        """
        Compute x**p in F_p^6 using square-and-multiply algorithm.
        Args:
        p: The exponent, a non-negative integer.
        Returns:
        x**p in F_p^6, represented similarly as x.
        """
        assert isinstance(p, int), f"Invalid exponent {p=}"

        # Handle the easy cases.
        if p == 0:
            # x**0 = 1, where 1 is the multiplicative identity in F_p^2.
            return self.one(self.curve_id)
        elif p == 1:
            # x**1 = x.
            return self
        elif p < 0:
            return self.__inv__() ** (-p)

        # Start the computation.
        result = self.one(
            self.curve_id
        )  # Initialize result as the multiplicative identity in F_p^2.
        temp = self  # Initialize temp as self.

        # Loop through each bit of the exponent p.
        for bit in reversed(bin(p)[2:]):  # [2:] to strip the "0b" prefix.
            if bit == "1":
                result = result * temp
            temp = temp * temp

        return result


@dataclass(slots=True, init=False)
class E12:
    c0: E6
    c1: E6
    curve_id: int

    def __init__(self, x: list[PyFelt | int | E6], curve_id: int):
        self.curve_id = curve_id
        if isinstance(x[0], (PyFelt, int)) and len(x) == 12:
            self.c0 = E6(x=x[0:6], curve_id=curve_id)
            self.c1 = E6(x=x[6:12], curve_id=curve_id)
        elif type(x[0] == E6 and len(x) == 2):
            self.c0 = x[0]
            self.c1 = x[1]
        else:
            raise ValueError

    def __hash__(self):
        return hash((tuple(self.value_coeffs), self.curve_id))

    @property
    def value_coeffs(self) -> list[int]:
        return [
            self.c0.b0.a0,
            self.c0.b0.a1,
            self.c0.b1.a0,
            self.c0.b1.a1,
            self.c0.b2.a0,
            self.c0.b2.a1,
            self.c1.b0.a0,
            self.c1.b0.a1,
            self.c1.b1.a0,
            self.c1.b1.a1,
            self.c1.b2.a0,
            self.c1.b2.a1,
        ]

    def print_as_sage_poly(self, var_name: str = "x"):
        field = get_base_field(self.curve_id)
        coeffs = [field(c) for c in self.value_coeffs]
        coeffs = tower_to_direct(coeffs, self.curve_id, 12)
        return Polynomial(coeffs).print_as_sage_poly(var_name)

    @staticmethod
    def from_poly(poly: Polynomial, curve_id: int):
        field = get_base_field(curve_id)
        coeffs = poly.get_coeffs()
        coeffs = coeffs + [field(0)] * (12 - len(coeffs))
        coeffs = direct_to_tower(coeffs, curve_id, 12)
        return E12(coeffs, curve_id)

    @staticmethod
    def from_direct(coeffs: list[PyFelt | ModuloCircuitElement], curve_id: int):
        coeffs = direct_to_tower([c.felt for c in coeffs], curve_id, 12)
        return E12(coeffs, curve_id)

    def to_poly(self) -> Polynomial:
        field = get_base_field(self.curve_id)
        coeffs = [field(c) for c in self.value_coeffs]
        coeffs = tower_to_direct(coeffs, self.curve_id, 12)
        return Polynomial(coeffs)

    def to_direct(self) -> list[PyFelt]:
        field = get_base_field(self.curve_id)
        coeffs = [field(c) for c in self.value_coeffs]
        coeffs = tower_to_direct(coeffs, self.curve_id, 12)
        return coeffs

    @property
    def order(self):
        return self.c0.b0.p**12

    @property
    def felt_coeffs(self) -> list[PyFelt]:
        return [PyFelt(c, self.c0.b0.p) for c in self.value_coeffs]

    @staticmethod
    def one(curve_id: int):
        return E12([E6.one(curve_id), E6.zero(curve_id)], curve_id)

    @staticmethod
    def zero(curve_id: int):
        return E12([E6.zero(curve_id), E6.zero(curve_id)], curve_id)

    @staticmethod
    def random(curve_id: int):
        field = get_base_field(curve_id)
        return E12([field(random.randint(0, field.p - 1)) for _ in range(12)], curve_id)

    def __mul__(self, other):
        a = self.c0 + self.c1
        b = other.c0 + other.c1
        a = a * b
        b = self.c0 * other.c0
        c = self.c1 * other.c1
        z1 = a - b - c
        c = c.mul_by_non_residue()
        z0 = c + b
        return E12([z0, z1], self.curve_id)

    def conjugate(self):
        return E12([self.c0, -self.c1], self.curve_id)

    def square(self):
        c0 = self.c0 - self.c1
        c3 = -(self.c1.mul_by_non_residue()) + self.c0
        c2 = self.c0 * self.c1
        c0 = (c0 * c3) + c2
        c1 = c2 + c2
        c2 = c2.mul_by_non_residue()
        c0 = c0 + c2
        return E12([c0, c1], self.curve_id)

    def __inv__(self):
        t0, t1 = self.c0 * self.c0, self.c1 * self.c1
        tmp = t1.mul_by_non_residue()
        t0 = t0 - tmp
        t1 = t0.__inv__()
        c0 = self.c0 * t1
        c1 = -self.c1 * t1
        return E12([c0, c1], curve_id=self.curve_id)

    def div(self, other):
        if isinstance(other, E12):
            return self * other.__inv__()
        raise NotImplementedError

    def __pow__(self, p: int):
        """
        Compute x**p in F_p^12 using square-and-multiply algorithm.
        Args:
        p: The exponent, a non-negative integer.
        Returns:
        x**p in F_p^12, represented similarly as x.
        """
        assert isinstance(p, int), f"Invalid exponent {p=}"

        # Handle the easy cases.
        if p == 0:
            # x**0 = 1, where 1 is the multiplicative identity in F_p^2.
            return self.one(self.curve_id)
        elif p == 1:
            # x**1 = x.
            return self
        elif p < 0:
            return self.__inv__() ** (-p)

        # Start the computation.
        result = self.one(
            self.curve_id
        )  # Initialize result as the multiplicative identity in F_p^2.
        temp = self  # Initialize temp as self.

        # Loop through each bit of the exponent p.
        for bit in reversed(bin(p)[2:]):  # [2:] to strip the "0b" prefix.
            if bit == "1":
                result = result * temp
            temp = temp.square()

        return result

    def final_exp(self, use_rust: bool = True):
        if use_rust:
            coeffs = garaga_rs.final_exp(self.curve_id, self.value_coeffs)
            return E12(coeffs, self.curve_id)
        else:
            cofactor = CURVES[self.curve_id].final_exp_cofactor
            h = (
                cofactor
                * (CURVES[self.curve_id].p ** 12 - 1)
                // CURVES[self.curve_id].n
            )
            return self**h

    def serialize(self) -> bytes:
        # Implement serialization like ark-ff:
        serialized = bytearray()
        bit_size = CURVES[self.curve_id].p.bit_length()
        byte_size = (bit_size + 7) // 8
        for c in self.value_coeffs[::-1]:
            serialized.extend(c.to_bytes(byte_size, byteorder="big"))

        return bytes(serialized)


def get_tower_object(x: list[PyFelt], curve_id: int, extension_degree: int):
    if extension_degree == 2:
        return E2(x[0].value, x[1].value, x[0].p)

    if extension_degree == 6:
        return E6(x, curve_id)
    elif extension_degree == 12:
        return E12(x, curve_id)
    else:
        raise ValueError



from math import gcd

from garaga.definitions import CURVES, CurveID
from garaga.hints.tower_backup import E12

"""
The goal of this script is to provide aux witness for efficient proving that Miller loop outputs of BLS12-381 curve are λ residues.
Co-factor h = 27 * p * h3 and λ = r * 3 * p^2.
We know that when pairings are equal multi Miller loop output is always r-th residue,
but it is λ residue only when it is a cube and p-th residue too.
This happens with a very small probability (1/3*p), but both 27th and p-th roots of unity are in the subfields of Fp12, Fp3 and Fp respectively.
Thus we can simply clear their contribution and not affect he soundness of the method.

Note that gcd(27*p, h3) = 1, therefore we do the following:

i := inverse_mod(h3, 27*p)
s := -i % 27*p

Then given the Miller loop output x, computing w = (x^h3)^s gives exactly the inverse of 27*p-th root of unity contribution in x.
By computing x_sh = w * x we are left with an element that is of order exactly h3.

Since gcd(h1, λ) = 1 computing λ root of x_sh can be simply done by just raising x_sh^e where e = inverse_mod(h3, λ)
"""

# Bls

x = CURVES[CurveID.BLS12_381.value].x
k = 12
r = x**4 - x**2 + 1
q = ((x - 1) ** 2) // 3 * r + x
h = (q**k - 1) // r

lam = -x + q
m = lam // r

p = 5044125407647214251
h3 = h // (27 * p)

e = pow(lam, -1, h3)
s = (-1 * pow(h3, -1, p * 27)) % (p * 27)


assert h % (27 * p) == 0
assert m == 3 * p**2

assert gcd(3, h3) == 1
assert gcd(p**2, h3) == 1
assert gcd(3, h3) == 1
assert gcd(p, h3) == 1
assert gcd(p, 27 * h3) == 1
assert gcd(27, p * h3) == 1

ONE = E12.one(curve_id=1)

assert (q**3 - 1) % 27 == 0


def get_root_and_scaling_factor_bls(mlo: E12) -> tuple[E12, E12]:
    """
    Takes a miller loop output and returns root, shift such that
    root**lam = shift * mlo, if and only if mlo**h == 1.
    """
    x = mlo**h3
    shift = x**s
    root = (shift * mlo) ** e
    return root, shift



import asyncio
import glob
import json
import os
import shutil
import subprocess
from enum import Enum
from importlib.metadata import PackageNotFoundError, version
from pathlib import Path

import rich
from starknet_py.contract import Contract
from starknet_py.net.account.account import Account
from starknet_py.net.client_errors import ClientError, ContractNotFoundError
from starknet_py.net.full_node_client import FullNodeClient
from starknet_py.net.models import StarknetChainId
from starknet_py.net.signer.stark_curve_signer import KeyPair

from garaga.definitions import ProofSystem
from garaga.hints.io import to_int


def get_package_version():
    try:
        __version__ = version("garaga")
    except PackageNotFoundError:
        # package is not installed
        __version__ = "dev"
    return __version__


class Network(Enum):
    SEPOLIA = "sepolia"
    MAINNET = "mainnet"


def load_account(network: Network):
    rpc_url = os.getenv(f"{network.name.upper()}_RPC_URL")
    account_address = os.getenv(f"{network.name.upper()}_ACCOUNT_ADDRESS")
    account_private_key = os.getenv(f"{network.name.upper()}_ACCOUNT_PRIVATE_KEY")
    if network == Network.MAINNET:
        chain = StarknetChainId.MAINNET
    else:
        chain = StarknetChainId.SEPOLIA

    client = FullNodeClient(node_url=rpc_url)
    account = Account(
        address=account_address,
        client=client,
        key_pair=KeyPair.from_private_key(to_int(account_private_key)),
        chain=chain,
    )
    account.ESTIMATED_AMOUNT_MULTIPLIER = 1.02
    account.ESTIMATED_FEE_MULTIPLIER = 1.02
    account.ESTIMATED_UNIT_PRICE_MULTIPLIER = 1.02
    return account


def get_contract_if_exists(account: Account, contract_address: int) -> Contract | None:
    try:
        res = asyncio.run(Contract.from_address(contract_address, account))
        return res
    except ContractNotFoundError:

        return None
    except ClientError as e:
        if "no contract with address" in e.message.lower():
            return None
        raise e


async def get_contract_if_exists_async(
    account: Account, contract_address: int
) -> Contract | None:
    try:
        res = await Contract.from_address(contract_address, account)
        return res
    except ContractNotFoundError:

        return None
    except ClientError as e:
        if "no contract with address" in e.message.lower():
            return None
        raise e


def get_contract_iff_exists(account: Account, contract_address: int) -> Contract:
    contract = get_contract_if_exists(account, contract_address)
    if contract is None:
        rich.print(
            f"[red]Contract {contract_address} does not exists on {account._chain_id.name}[/red]"
        )
        raise ValueError(f"Contract {contract_address} not found")
    return contract


def create_directory(path: str):
    if not os.path.exists(path):
        os.makedirs(path)
        # print(f"Directory created: {path}")


def complete_pairing_curve_id(incomplete: str):
    curve_ids = ["0", "1"]  # Corresponding to BN254 and BLS12_381
    return [cid for cid in curve_ids if cid.startswith(incomplete)]


def complete_proof_system(incomplete: str):
    systems = [ps.value for ps in ProofSystem]
    return [system for system in systems if system.startswith(incomplete)]


def complete_network(incomplete: str):
    networks = [network.name for network in StarknetChainId]
    return [network for network in networks if network.startswith(incomplete)]


def get_voyager_network_prefix(network: Network) -> str:
    return "" if network == Network.MAINNET else "sepolia."


def voyager_link_tx(network: Network, tx_hash: int) -> str:
    voyager_prefix = get_voyager_network_prefix(network)
    return f"https://{voyager_prefix}voyager.online/tx/{hex(tx_hash)}"


def voyager_link_class(network: Network, class_hash: int) -> str:
    voyager_prefix = get_voyager_network_prefix(network)
    return f"https://{voyager_prefix}voyager.online/class/{hex(class_hash)}"


def scarb_build_contract_folder(contract_folder_path: str):
    cmd = "scarb build"
    print(f"Running command: {cmd} in directory: {contract_folder_path}")
    try:
        subprocess.run(
            cmd,
            cwd=contract_folder_path,
            check=True,
            shell=True,
            capture_output=True,
            text=True,
        )
    except subprocess.CalledProcessError as e:
        print(f"Error building contract: {e}")
        print(f"Command output:\n{e.output}")
        raise
    except FileNotFoundError as e:
        print(f"Error: {e}")
        raise
    except Exception as e:
        print(f"An unexpected error occurred: {e}")
        raise


def get_sierra_casm_artifacts(
    contract_folder_path: str,
) -> tuple[None, None] | tuple[str, str]:
    """
    Get the Sierra and CASM artifacts for a contract.
    """
    target_dir = os.path.join(contract_folder_path, "target/dev/")

    # Clean the target/dev/ folder if it already exists
    if os.path.exists(target_dir):
        shutil.rmtree(target_dir)
    os.makedirs(target_dir)

    scarb_build_contract_folder(contract_folder_path)

    artifacts_file = glob.glob(os.path.join(target_dir, "*.starknet_artifacts.json"))
    assert (
        len(artifacts_file) == 1
    ), "Artifacts JSON file not found or multiple files found."

    with open(artifacts_file[0], "r") as f:
        artifacts = json.load(f)

    contracts = artifacts["contracts"]
    if len(contracts) == 0:
        return None, None

    sierra_file = contracts[0]["artifacts"]["sierra"]
    casm_file = contracts[0]["artifacts"]["casm"]

    sierra_path = os.path.join(target_dir, sierra_file)
    casm_path = os.path.join(target_dir, casm_file)

    assert os.path.exists(sierra_path), f"Sierra file not found: {sierra_path}"
    assert os.path.exists(casm_path), f"CASM file not found: {casm_path}"

    with open(sierra_path, "r") as f:
        sierra = f.read()
    with open(casm_path, "r") as f:
        casm = f.read()
    return sierra, casm


def get_default_vk_path(system: ProofSystem) -> Path:
    script_path = Path(__file__).resolve()
    script_folder_path = script_path.parent
    match system:
        case ProofSystem.Risc0Groth16:
            return (
                script_folder_path.parent
                / "groth16_contract_generator"
                / "examples"
                / "vk_risc0.json"
            )
        case _:
            raise ValueError(
                f"Proof system {system} requires an user-provided verification key"
            )



import random
from dataclasses import dataclass
from functools import lru_cache
from hashlib import sha512

from garaga import garaga_rs
from garaga.definitions import (
    BASE,
    CURVES,
    N_LIMBS,
    CurveID,
    G1Point,
    TwistedEdwardsCurve,
)
from garaga.hints.io import bigint_split, split_128


@dataclass(slots=True)
class SchnorrSignature:
    """
    A Schnorr signature with associated public key and challenge.
    Fields :
    rx : int - The x-coordinate of the R point from the signature.
    s : int - The s-coordinate of the signature.
    e : int - The challenge hash.
    px : int - The x-coordinate of the public key.
    py : int - The y-coordinate of the public key.
    curve_id : CurveID - The curve id.
    """

    rx: int
    s: int
    e: int
    px: int
    py: int
    curve_id: CurveID

    def __post_init__(self):
        assert self.py % 2 == 0, "y-coordinate of public key must be even"
        n = CURVES[self.curve_id.value].n
        p = CURVES[self.curve_id.value].p
        assert (
            0 <= self.rx < p
        ), f"rx must be in range 0 < rx < {hex(p)}, got {hex(self.rx)}"
        assert 0 < self.s < n, f"s must be in range 0 < s < {hex(n)}, got {hex(self.s)}"
        assert 0 < self.e < n, f"e must be in range 0 < e < {hex(n)}, got {hex(self.e)}"

    @classmethod
    def sample(cls, curve_id: CurveID) -> "SchnorrSignature":
        """
        Generate a valid Schnorr signature for the given curve.
        The algorithm:
        1. Generate private key x
        2. Compute public key P = xG
        3. Generate random k (nonce)
        4. Compute R = kG
        5. Compute e (challenge)
        6. Compute s = k + ex (mod n)
        """
        curve = CURVES[curve_id.value]
        n = curve.n
        G = G1Point.get_nG(curve_id, 1)

        # Generate private key and compute public key
        x = random.randint(1, n - 1)
        Pk = G.scalar_mul(x)

        # Ensure public key has even y (BIP340 requirement)
        if Pk.y % 2 == 1:
            x = (-x) % n  # Update the private key accordingly
            Pk = -Pk

        # Generate nonce and compute R point
        k = random.randint(1, n - 1)
        R = G.scalar_mul(k)

        # Ensure R has even y (BIP340 requirement)
        if R.y % 2 == 1:
            k = (-k) % n  # Update the nonce accordingly
            R = -R

        # Generate challenge (in real usage this would be H(R.x || P.x || message))
        e = random.randint(1, n - 1)

        # Compute s = k + ex mod n
        s = (k + e * x) % n

        res = SchnorrSignature(rx=R.x, s=s, e=e, px=Pk.x, py=Pk.y, curve_id=curve_id)
        assert res.is_valid(), "generated signature is invalid"
        return res

    def is_valid(self) -> bool:
        Pk = G1Point(self.px, self.py, self.curve_id)
        n = CURVES[self.curve_id.value].n
        G = G1Point.get_nG(self.curve_id, 1)
        res = G.scalar_mul(self.s).add(Pk.scalar_mul(-self.e % n))
        return res.x == self.rx

    def serialize(self) -> list[int]:
        cd = []
        cd.extend(bigint_split(self.rx, N_LIMBS, BASE))
        cd.extend(split_128(self.s))
        cd.extend(split_128(self.e))
        cd.extend(bigint_split(self.px, N_LIMBS, BASE))
        cd.extend(bigint_split(self.py, N_LIMBS, BASE))
        return cd

    def serialize_with_hints(self, use_rust=False, as_str=False) -> list[int] | str:
        """Serialize the signature with hints for verification"""
        if use_rust:
            cd = garaga_rs.schnorr_calldata_builder(
                self.rx, self.s, self.e, self.px, self.py, self.curve_id.value
            )
            if as_str:
                return "[{}]".format(", ".join(map(hex, cd)))
            return cd

        cd = self.serialize()
        e_neg = -self.e % CURVES[self.curve_id.value].n
        msm_calldata = garaga_rs.msm_calldata_builder(
            [
                CURVES[self.curve_id.value].Gx,
                CURVES[self.curve_id.value].Gy,
                self.px,
                self.py,
            ],
            [self.s, e_neg],
            self.curve_id.value,
            False,  # include_points_and_scalars
            True,  # serialize_as_pure_felt252_array
        )
        cd.extend(msm_calldata)
        if as_str:
            return "[{}]".format(", ".join(map(hex, cd)))
        return cd


@dataclass(slots=True)
class ECDSASignature:
    """
    An ECDSA signature with recovery parameter and associated public key and message hash.

    Fields:
      r    : int - r component of the signature (derived from the ephemeral point R.x mod n).
      s    : int - s component of the signature.
      v    : int - Recovery parameter; 0 if R.y is even, otherwise 1.
      px   : int - The x-coordinate of the public key.
      py   : int - The y-coordinate of the public key.
      z    : int - The message hash used in signing (in practice, z = H(message)).
      curve_id : CurveID - The identifier for the curve being used.
    """

    r: int
    s: int
    v: int
    px: int
    py: int
    z: int
    curve_id: CurveID

    def __post_init__(self):
        curve = CURVES[self.curve_id.value]
        n = curve.n
        p = curve.p
        # For ECDSA, r and s must be nonzero and less than n.
        assert 0 < self.r < n, f"r must be in range 1..n-1, got {hex(self.r)}"
        assert 0 < self.s < n, f"s must be in range 1..n-1, got {hex(self.s)}"
        # v should be either 0 or 1.
        assert self.v in (0, 1), f"v must be 0 or 1, got {self.v}"
        # Ensure public key coordinates are in the field.
        assert 0 <= self.px < p, f"px must be in range 0..p-1, got {hex(self.px)}"
        assert 0 <= self.py < p, f"py must be in range 0..p-1, got {hex(self.py)}"

    @classmethod
    def sample(cls, curve_id: CurveID) -> "ECDSASignature":
        """
        Generate a valid ECDSA signature for the given curve.

        The algorithm:
          1. Generate a private key d.
          2. Compute the public key Q = d * G.
          3. Generate a random message hash z (in practice, z = H(message)).
          4. Choose a random nonce k and compute R = k * G.
          5. Let r = R.x mod n (with r != 0).
          6. Compute s = k⁻¹ * (z + r * d) mod n (with s != 0).
          7. Set the recovery parameter v = 0 if R.y is even, else 1.
        """
        curve = CURVES[curve_id.value]
        n = curve.n
        p = curve.p
        G = G1Point.get_nG(curve_id, 1)

        # Generate the private key d and public key Q.
        d = random.randint(1, n - 1)
        Q = G.scalar_mul(d)

        # Generate a random message hash (normally, this is H(message)).
        z = random.randint(1, n - 1)

        # Choose a nonce k until we get valid r and s.
        while True:
            k = random.randint(1, n - 1)
            R = G.scalar_mul(k)
            r = R.x % n  # r is computed from the x-coordinate of R.
            if r == 0:
                continue
            try:
                k_inv = pow(k, -1, n)
            except ValueError:
                continue
            s = (k_inv * (z + r * d)) % n
            if s == 0:
                continue
            v = 0 if R.y % 2 == 0 else 1
            break

        sig = cls(r=r, s=s, v=v, px=Q.x, py=Q.y, z=z, curve_id=curve_id)
        assert (
            sig.is_valid()
        ), f"generated ECDSA signature on curve {curve_id.name} is invalid"
        return sig

    def is_valid(self) -> bool:
        """
        Verify the ECDSA signature using the stored message hash and public key.

        Standard verification:
          1. Compute w = s⁻¹ mod n.
          2. Compute u₁ = z * w mod n and u₂ = r * w mod n.
          3. Compute R' = u₁ * G + u₂ * Q.
          4. The signature is valid if (R'.x mod n) equals r.
        """
        curve = CURVES[self.curve_id.value]
        n = curve.n
        G = G1Point.get_nG(self.curve_id, 1)
        Q = G1Point(self.px, self.py, self.curve_id)

        try:
            s_inv = pow(self.s, -1, n)
        except ValueError:
            return False

        u1 = (self.z * s_inv) % n
        u2 = (self.r * s_inv) % n

        R_prime = G.scalar_mul(u1).add(Q.scalar_mul(u2))
        return (R_prime.x % n) == self.r
        # return (R_prime.x) == self.r

    def serialize(self) -> list[int]:
        cd = []
        cd.extend(bigint_split(self.r, N_LIMBS, BASE))
        cd.extend(split_128(self.s))
        cd.append(self.v)
        cd.extend(bigint_split(self.px, N_LIMBS, BASE))
        cd.extend(bigint_split(self.py, N_LIMBS, BASE))
        cd.extend(split_128(self.z))
        return cd

    def serialize_with_hints(self, use_rust=False, as_str=False) -> list[int] | str:
        """Serialize the signature with hints for verification"""
        if use_rust:
            cd = garaga_rs.ecdsa_calldata_builder(
                self.r, self.s, self.v, self.px, self.py, self.z, self.curve_id.value
            )
            if as_str:
                return "[{}]".format(", ".join(map(hex, cd)))
            return cd

        cd = self.serialize()
        n = CURVES[self.curve_id.value].n
        s_inv = pow(self.s, -1, n)
        u1 = (self.z * s_inv) % n
        u2 = (self.r * s_inv) % n

        msm_calldata = garaga_rs.msm_calldata_builder(
            [
                CURVES[self.curve_id.value].Gx,
                CURVES[self.curve_id.value].Gy,
                self.px,
                self.py,
            ],
            [u1, u2],
            self.curve_id.value,
            False,  # include_points_and_scalars
            True,  # serialize_as_pure_felt252_array
        )
        cd.extend(msm_calldata)
        if as_str:
            return "[{}]".format(", ".join(map(hex, cd)))
        return cd


@dataclass(slots=True)
class EdDSA25519Signature:
    """
    An EdDSA25519 signature with associated public key and message.
    """

    Ry_twisted: int
    s: int
    Py_twisted: int
    msg: bytes

    def __eq__(self, other) -> bool:
        if not isinstance(other, EdDSA25519Signature):
            return NotImplemented
        return (
            self.Ry_twisted == other.Ry_twisted
            and self.s == other.s
            and self.Py_twisted == other.Py_twisted
            and self.msg == other.msg
        )

    def __hash__(self) -> int:
        return hash((self.Ry_twisted, self.s, self.Py_twisted, self.msg))

    def __post_init__(self):
        assert 0 <= self.Ry_twisted < 2**256
        assert 0 <= self.Py_twisted < 2**256
        assert 0 <= self.s < 2**256

    @property
    def curve_id(self) -> CurveID:
        return CurveID.ED25519

    @property
    def curve(self) -> TwistedEdwardsCurve:
        return CURVES[CurveID.ED25519.value]

    @classmethod
    def from_json(cls, json):
        public_key_bytes = bytes.fromhex(json["public_key"])
        message_bytes = bytes.fromhex(json["message"])
        signature_bytes = bytes.fromhex(json["signature"])

        assert len(public_key_bytes) == 32
        assert len(signature_bytes) == 64

        return cls(
            Ry_twisted=int.from_bytes(signature_bytes[:32], "little"),
            s=int.from_bytes(signature_bytes[32:64], "little"),
            Py_twisted=int.from_bytes(public_key_bytes, "little"),
            msg=message_bytes,
        )

    def serialize(self) -> list[int]:
        cd = []
        cd.extend(split_128(self.Ry_twisted))
        cd.extend(split_128(self.s))
        cd.extend(split_128(self.Py_twisted))
        cd.append(len(self.msg))
        cd.extend(list(self.msg))

        return cd

    def xrecover(self, y: int) -> int:
        p = self.curve.p
        D = self.curve.d_twisted
        I = pow(2, (p - 1) // 4, p)
        xx = (y * y - 1) * pow(D * y * y + 1, -1, p)
        x = pow(xx, (p + 3) // 8, p)
        if (x * x - xx) % p != 0:
            x = (x * I) % p
            assert (x * x - xx) % p == 0
        if x % 2 != 0:
            x = p - x
        return x

    def decode_point(self, compressed_point_le: int) -> G1Point:
        sign_bit, y_twisted = divmod(compressed_point_le, 2 ** (255))
        x_twisted = self.xrecover(y_twisted)
        if x_twisted % 2 != sign_bit:
            x_twisted = self.curve.p - x_twisted
        x_weierstrass, y_weierstrass = self.curve.to_weierstrass(x_twisted, y_twisted)
        return G1Point(x_weierstrass, y_weierstrass, self.curve_id)

    @lru_cache(maxsize=None)
    def deserialize_R_A_h(self) -> tuple[G1Point, G1Point, int]:
        R = self.decode_point(self.Ry_twisted)
        A = self.decode_point(self.Py_twisted)

        preimage = (
            self.Ry_twisted.to_bytes(32, "little")
            + self.Py_twisted.to_bytes(32, "little")
            + self.msg
        )
        H_bytes = sha512(preimage).digest()
        H = int.from_bytes(H_bytes, "little")
        h = H % self.curve.n
        return R, A, h

    def is_valid(self) -> bool:
        s = self.s

        R, A, h = self.deserialize_R_A_h()
        G_neg = G1Point.get_nG(self.curve_id, -1)

        msm_result = G1Point.msm(points=[G_neg, A], scalars=[s, h])

        should_be_inf = msm_result.add(R)

        if R.is_infinity() or A.is_infinity():
            return False

        return should_be_inf.is_infinity()

    def serialize_with_hints(self, use_rust=False, as_str=False) -> list[int] | str:
        """Serialize the signature with hints for verification"""
        if use_rust:
            return garaga_rs.eddsa_calldata_builder(
                self.Ry_twisted,
                self.s,
                self.Py_twisted,
                self.msg,
            )

        cd = self.serialize()
        R, A, h = self.deserialize_R_A_h()
        msm_calldata = garaga_rs.msm_calldata_builder(
            [
                CURVES[self.curve_id.value].Gx,
                -CURVES[self.curve_id.value].Gy % self.curve.p,
                A.x,
                A.y,
            ],
            [self.s, h],
            self.curve_id.value,
            False,  # include_points_and_scalars
            True,  # serialize_as_pure_felt252_array
        )

        cd.extend(msm_calldata)
        (Rx_twisted, _) = self.curve.to_twistededwards(R.x, R.y)
        cd.extend(split_128(Rx_twisted))  # sqrt_Rx_hint
        (Px_twisted, _) = self.curve.to_twistededwards(A.x, A.y)
        cd.extend(split_128(Px_twisted))  # sqrt_Px_hint

        if as_str:
            return "[{}]".format(", ".join(map(hex, cd)))
        return cd

    def __eq__(self, other) -> bool:
        if not isinstance(other, EdDSA25519Signature):
            return NotImplemented
        return (
            self.Ry_twisted == other.Ry_twisted
            and self.s == other.s
            and self.Py_twisted == other.Py_twisted
            and self.msg == other.msg
        )

    def __hash__(self) -> int:
        return hash((self.Ry_twisted, self.s, self.Py_twisted, self.msg))


if __name__ == "__main__":
    import json

    with open("tests/ed25519_test_vectors.json", "r") as f:
        test_vectors = json.load(f)

    for i, test_vector in enumerate(test_vectors):
        signature = EdDSA25519Signature.from_json(test_vector)
        assert signature.is_valid(), f"Signature {i} is invalid"



from typing import Iterator

from garaga.definitions import BLS12_381_ID, BN254_ID, CURVES
from garaga.extension_field_modulo_circuit import ModuloCircuitElement
from garaga.precompiled_circuits.multi_miller_loop import MultiMillerLoopCircuit


def tower_line_function_sparsity(curve_id: int) -> list[int]:
    match curve_id:
        case 0:
            return [2, 0, 0, 0, 0, 0, 1, 1, 1, 1, 0, 0]
        case 1:
            return [1, 1, 1, 1, 0, 0, 0, 0, 2, 0, 0, 0]
        case _:
            raise ValueError(f"Invalid curve ID: {curve_id}")


class MillerTowerCircuit(MultiMillerLoopCircuit):
    def __init__(
        self,
        name: str,
        curve_id: int,
        n_pairs: int,
        compilation_mode: int = 1,
    ):
        super().__init__(
            name=name,
            curve_id=curve_id,
            n_pairs=n_pairs,
            compilation_mode=compilation_mode,
        )

    def eval_tower_line(self, line: Iterator[ModuloCircuitElement], yInv, xNegOverY):
        assert len(line) == 4
        # Input is R0 = slope, R1 = intercept
        r0 = [line[0], line[1]]
        r1 = [line[2], line[3]]
        if self.curve_id == BN254_ID:
            return [
                self.mul(r0[0], xNegOverY),
                self.mul(r0[1], xNegOverY),
                self.mul(r1[0], yInv),
                self.mul(r1[1], yInv),
            ]
        elif self.curve_id == BLS12_381_ID:
            return [
                self.mul(r1[0], yInv),
                self.mul(r1[1], yInv),
                self.mul(r0[0], xNegOverY),
                self.mul(r0[1], xNegOverY),
            ]
        else:
            raise NotImplementedError

    def mul_by_line_tower(
        self,
        tower_fp12: list[ModuloCircuitElement],
        line_evaluated: list[ModuloCircuitElement],
    ):
        assert len(tower_fp12) == 12
        assert (
            len(line_evaluated) == 4
        ), f"Expected 4 elements, got {len(line_evaluated)}"
        c0_fp6, c1_fp6 = tower_fp12[0:6], tower_fp12[6:12]
        # c0_b0, c0_b1, c0_b2 = c0[0:2], c0[2:4], c0[4:6]
        # c1_b0, c1_b1, c1_b2 = c1[0:2], c1[2:4], c1[4:6]

        # MulBy34. multiplication by sparse element (1,0,0,c3,c4,0)
        if self.curve_id == 0:
            a = c0_fp6
            b = c1_fp6
            c3, c4 = line_evaluated[0:2], line_evaluated[2:4]
            b = self._fp6_by_01(b, c3, c4)

            d0 = self.vector_add(
                [self.set_or_get_constant(1), self.set_or_get_constant(0)], c3
            )
            d = self.vector_add(c0_fp6, c1_fp6)
            d = self._fp6_by_01(d, d0, c4)

            z_c1 = self.vector_add(self.vector_neg(self.vector_add(a, b)), d)
            z_c0 = self.vector_add(self.fp6_mul_by_non_residue(b), a)

            return z_c0 + z_c1
        elif self.curve_id == 1:
            # Mul By 01
            # // MulBy01 multiplication by sparse element (c0, c1, 0, 0, 1)
            # func (z *E12) MulBy01(c0, c1 *E2) *E12 {

            # 	var a, b E6
            # 	var d E2

            # 	a.Set(&z.C0)
            # 	a.MulBy01(c0, c1)

            # 	b.MulByNonResidue(&z.C1)
            # 	d.SetOne().Add(c1, &d)

            # 	z.C1.Add(&z.C1, &z.C0)
            # 	z.C1.MulBy01(c0, &d)
            # 	z.C1.Sub(&z.C1, &a)
            # 	z.C1.Sub(&z.C1, &b)
            # 	z.C0.MulByNonResidue(&b)
            # 	z.C0.Add(&z.C0, &a)

            # 	return z
            # }
            c0, c1 = line_evaluated[0:2], line_evaluated[2:4]

            a = c0_fp6
            a = self._fp6_by_01(a, c0, c1)
            b = self.fp6_mul_by_non_residue(c1_fp6)

            d = [self.add(self.set_or_get_constant(1), c1[0]), c1[1]]

            z_c1 = self.vector_add(c1_fp6, c0_fp6)
            z_c1 = self._fp6_by_01(z_c1, c0, d)
            z_c1 = self.vector_sub(z_c1, a)
            z_c1 = self.vector_sub(z_c1, b)

            z_c0 = self.fp6_mul_by_non_residue(b)
            z_c0 = self.vector_add(z_c0, a)
            return z_c0 + z_c1

    def mul_by_line_line_tower(self, tower_fp12, line_line):
        assert len(tower_fp12) == 12
        assert len(line_line) == 10
        c0_fp6, c1_fp6 = tower_fp12[0:6], tower_fp12[6:12]

    def mul_line_by_line_tower(self, l1, l2):
        assert len(l1) == 4
        assert len(l2) == 4
        if self.curve_id == 0:
            d3, d4 = l1[0:2], l1[2:4]
            c3, c4 = l2[0:2], l2[2:4]

            x3 = self.fp2_mul(c3, d3)
            x4 = self.fp2_mul(c4, d4)
            x04 = self.vector_add(c4, d4)
            x03 = self.vector_add(c3, d3)
            tmp = self.vector_add(c3, c4)
            x34 = self.vector_sub(
                self.vector_sub(self.fp2_mul(self.vector_add(d3, d4), tmp), x3), x4
            )
            z00 = self.fp2_mul_by_non_residue(x4)
            z00 = [self.add(z00[0], self.set_or_get_constant(1)), z00[1]]

            return z00 + x3 + x34 + x03 + x04
        if self.curve_id == 1:

            c0, c1 = l2[0:2], l2[2:4]
            d0, d1 = l1[0:2], l1[2:4]

            x0 = self.fp2_mul(c0, d0)
            x1 = self.fp2_mul(c1, d1)

            x04 = self.vector_add(d0, c0)
            tmp = self.vector_add(c0, c1)
            x01 = self.vector_sub(
                self.vector_sub(self.fp2_mul(self.vector_add(d0, d1), tmp), x0), x1
            )
            x14 = self.vector_add(d1, c1)
            z00 = self.fp2_mul_by_non_residue(
                [self.set_or_get_constant(1), self.set_or_get_constant(0)]
            )
            z00 = self.vector_add(z00, x0)

            # C0B0 + C0B1 + C0B2 + C1B1 + C1B2
            return z00 + x01 + x1 + x04 + x14

    def _fp6_by_01(self, e6: list[ModuloCircuitElement], c0, c1):
        assert len(e6) == 6
        assert len(c0) == 2
        assert len(c1) == 2
        b0, b1, b2 = e6[0:2], e6[2:4], e6[4:6]
        a = self.fp2_mul(b0, c0)
        b = self.fp2_mul(b1, c1)
        tmp = self.vector_add(b1, b2)
        t0 = self.fp2_mul(c1, tmp)
        t0 = self.vector_sub(t0, b)
        t0 = self.fp2_mul_by_non_residue(t0)
        t0 = self.vector_add(t0, a)

        tmp = self.vector_add(b0, b2)
        t2 = self.fp2_mul(c0, tmp)
        t2 = self.vector_sub(t2, a)
        t2 = self.vector_add(t2, b)

        t1 = self.vector_add(c0, c1)
        tmp = self.vector_add(b0, b1)
        t1 = self.fp2_mul(t1, tmp)
        t1 = self.vector_sub(t1, a)
        t1 = self.vector_sub(t1, b)

        b0 = t0
        b1 = t1
        b2 = t2
        # flatten lists:
        return b0 + b1 + b2

    def fp6_mul_by_non_residue(self, a: list[ModuloCircuitElement]):
        assert len(a) == 6
        a = list(a)
        b0, b1, b2 = a[0:2], a[2:4], a[4:6]
        return self.fp2_mul_by_non_residue(b2) + b0 + b1

    def fp12_square(self, a: list[ModuloCircuitElement]):
        assert len(a) == 12

        xc0, xc1 = a[0:6], a[6:12]
        c0 = self.vector_sub(xc0, xc1)
        c3 = self.vector_add(xc0, self.vector_neg(self.fp6_mul_by_non_residue(xc1)))
        c2 = self.fp6_mul(xc0, xc1)
        c0 = self.vector_add(self.fp6_mul(c0, c3), c2)
        z_c1 = self.vector_add(c2, c2)
        z_c0 = self.vector_add(c0, self.fp6_mul_by_non_residue(c2))

        return z_c0 + z_c1

    def fp12_mul(self, x: list[ModuloCircuitElement], y: list[ModuloCircuitElement]):
        assert len(x) == 12
        assert len(y) == 12

        x_c0, x_c1 = x[0:6], x[6:12]
        y_c0, y_c1 = y[0:6], y[6:12]

        a = self.vector_add(x_c0, x_c1)
        b = self.vector_add(y_c0, y_c1)
        a = self.fp6_mul(a, b)
        b = self.fp6_mul(x_c0, y_c0)
        c = self.fp6_mul(x_c1, y_c1)
        z_c1 = self.vector_sub(a, b)
        z_c1 = self.vector_sub(z_c1, c)
        z_c0 = self.fp6_mul_by_non_residue(c)
        z_c0 = self.vector_add(z_c0, b)

        return z_c0 + z_c1

    def fp6_mul(self, a: list[ModuloCircuitElement], b: list[ModuloCircuitElement]):
        assert len(a) == 6
        assert len(b) == 6

        xb0, xb1, xb2 = a[0:2], a[2:4], a[4:6]
        yb0, yb1, yb2 = b[0:2], b[2:4], b[4:6]

        t0 = self.fp2_mul(xb0, yb0)
        t1 = self.fp2_mul(xb1, yb1)
        t2 = self.fp2_mul(xb2, yb2)

        c0 = self.vector_add(xb1, xb2)
        tmp = self.vector_add(yb1, yb2)

        c0 = self.vector_add(
            self.fp2_mul_by_non_residue(
                self.vector_sub(self.vector_sub(self.fp2_mul(c0, tmp), t1), t2)
            ),
            t0,
        )

        c1 = self.vector_add(xb0, xb1)
        tmp = self.vector_add(yb0, yb1)
        c1 = self.vector_sub(self.vector_sub(self.fp2_mul(c1, tmp), t0), t1)
        tmp = self.fp2_mul_by_non_residue(t2)
        c1 = self.vector_add(c1, tmp)

        tmp = self.vector_add(xb0, xb2)
        c2 = self.vector_add(
            self.vector_sub(
                self.vector_sub(self.fp2_mul(self.vector_add(yb0, yb2), tmp), t0), t2
            ),
            t1,
        )

        return c0 + c1 + c2

    def fp12_inverse(self, a: list[ModuloCircuitElement]):

        assert len(a) == 12

        xc0, xc1 = a[0:6], a[6:12]

        t0 = self.fp6_square(xc0)
        t1 = self.fp6_square(xc1)
        tmp = self.fp6_mul_by_non_residue(t1)
        t0 = self.vector_sub(t0, tmp)
        t1 = self.fp6_inverse(t0)
        z_c0 = self.fp6_mul(xc0, t1)
        z_c1 = self.fp6_mul(xc1, t1)
        z_c1 = self.vector_neg(z_c1)

        return z_c0 + z_c1

    def fp6_inverse(self, a: list[ModuloCircuitElement]):
        assert len(a) == 6
        # func (z *E6) Inverse(x *E6) *E6 {
        # 	// Algorithm 17 from https://eprint.iacr.org/2010/354.pdf
        # 	// step 9 is wrong in the paper it's t1-t4
        # 	var t0, t1, t2, t3, t4, t5, t6, c0, c1, c2, d1, d2 E2
        # 	t0.Square(&x.B0)
        # 	t1.Square(&x.B1)
        # 	t2.Square(&x.B2)
        # 	t3.Mul(&x.B0, &x.B1)
        # 	t4.Mul(&x.B0, &x.B2)
        # 	t5.Mul(&x.B1, &x.B2)
        # 	c0.MulByNonResidue(&t5).Neg(&c0).Add(&c0, &t0)
        # 	c1.MulByNonResidue(&t2).Sub(&c1, &t3)
        # 	c2.Sub(&t1, &t4)
        # 	t6.Mul(&x.B0, &c0)
        # 	d1.Mul(&x.B2, &c1)
        # 	d2.Mul(&x.B1, &c2)
        # 	d1.Add(&d1, &d2).MulByNonResidue(&d1)
        # 	t6.Add(&t6, &d1)
        # 	t6.Inverse(&t6)
        # 	z.B0.Mul(&c0, &t6)
        # 	z.B1.Mul(&c1, &t6)
        # 	z.B2.Mul(&c2, &t6)
        xb0, xb1, xb2 = a[0:2], a[2:4], a[4:6]
        t0 = self.fp2_square(xb0)
        t1 = self.fp2_square(xb1)
        t2 = self.fp2_square(xb2)

        t3 = self.fp2_mul(xb0, xb1)
        t4 = self.fp2_mul(xb0, xb2)
        t5 = self.fp2_mul(xb1, xb2)
        c0 = self.vector_add(self.vector_neg(self.fp2_mul_by_non_residue(t5)), t0)
        c1 = self.vector_sub(self.fp2_mul_by_non_residue(t2), t3)
        c2 = self.vector_sub(t1, t4)

        t6 = self.fp2_mul(xb0, c0)
        d1 = self.fp2_mul(xb2, c1)
        d2 = self.fp2_mul(xb1, c2)
        d1 = self.fp2_mul_by_non_residue(self.vector_add(d1, d2))
        t6 = self.vector_add(t6, d1)
        t6 = self.fp2_inv(t6)
        zb0 = self.fp2_mul(c0, t6)
        zb1 = self.fp2_mul(c1, t6)
        zb2 = self.fp2_mul(c2, t6)

        return zb0 + zb1 + zb2

    def fp6_square(self, a: list[ModuloCircuitElement]):
        assert len(a) == 6

        xb0, xb1, xb2 = a[0:2], a[2:4], a[4:6]
        c4 = self.fp2_mul(xb0, xb1)
        c4 = self.vector_add(c4, c4)
        c5 = self.fp2_square(xb2)
        c1 = self.vector_add(self.fp2_mul_by_non_residue(c5), c4)
        c2 = self.vector_sub(c4, c5)
        c3 = self.fp2_square(xb0)
        c4 = self.vector_add(self.vector_sub(xb0, xb1), xb2)
        c5 = self.fp2_mul(xb1, xb2)
        c5 = self.vector_add(c5, c5)

        c4 = self.fp2_square(c4)
        c0 = self.vector_add(self.fp2_mul_by_non_residue(c5), c3)
        zb2 = self.vector_sub(self.vector_add(self.vector_add(c2, c4), c5), c3)
        zb0 = c0
        zb1 = c1

        return zb0 + zb1 + zb2

    def fp12_conjugate(self, a: list[ModuloCircuitElement]):
        assert len(a) == 12

        raise NotImplementedError("Not implemented")

    def fp12_cyclotomic_square(self, a: list[ModuloCircuitElement]):
        assert len(a) == 12

        xc0, xc1 = a[0:6], a[6:12]
        xc0b0 = xc0[0:2]
        xc0b1 = xc0[2:4]
        xc0b2 = xc0[4:6]
        xc1b0 = xc1[0:2]
        xc1b1 = xc1[2:4]
        xc1b2 = xc1[4:6]

        t0 = self.fp2_square(xc1b1)
        t1 = self.fp2_square(xc0b0)
        t6 = self.vector_add(xc1b1, xc0b0)
        t6 = self.fp2_square(t6)
        t6 = self.vector_sub(t6, t0)
        t6 = self.vector_sub(t6, t1)
        t2 = self.fp2_square(xc0b2)
        t3 = self.fp2_square(xc1b0)
        t7 = self.vector_add(xc0b2, xc1b0)
        t7 = self.fp2_square(t7)
        t7 = self.vector_sub(t7, t2)
        t7 = self.vector_sub(t7, t3)

        t4 = self.fp2_square(xc1b2)
        t5 = self.fp2_square(xc0b1)
        t8 = self.vector_add(xc1b2, xc0b1)
        t8 = self.fp2_square(t8)
        t8 = self.vector_sub(t8, t4)
        t8 = self.vector_sub(t8, t5)
        t8 = self.fp2_mul_by_non_residue(t8)

        t0 = self.fp2_mul_by_non_residue(t0)
        t0 = self.vector_add(t0, t1)
        t2 = self.fp2_mul_by_non_residue(t2)
        t2 = self.vector_add(t2, t3)
        t4 = self.fp2_mul_by_non_residue(t4)
        t4 = self.vector_add(t4, t5)

        zc0b0 = self.vector_sub(t0, xc0b0)
        zc0b0 = self.vector_add(zc0b0, zc0b0)
        zc0b0 = self.vector_add(zc0b0, t0)

        zc0b1 = self.vector_sub(t2, xc0b1)
        zc0b1 = self.vector_add(zc0b1, zc0b1)
        zc0b1 = self.vector_add(zc0b1, t2)

        zc0b2 = self.vector_sub(t4, xc0b2)
        zc0b2 = self.vector_add(zc0b2, zc0b2)
        zc0b2 = self.vector_add(zc0b2, t4)

        zc1b0 = self.vector_add(t8, xc1b0)
        zc1b0 = self.vector_add(zc1b0, zc1b0)
        zc1b0 = self.vector_add(zc1b0, t8)

        zc1b1 = self.vector_add(t6, xc1b1)
        zc1b1 = self.vector_add(zc1b1, zc1b1)
        zc1b1 = self.vector_add(zc1b1, t6)

        zc1b2 = self.vector_add(t7, xc1b2)
        zc1b2 = self.vector_add(zc1b2, zc1b2)
        zc1b2 = self.vector_add(zc1b2, t7)

        return zc0b0 + zc0b1 + zc0b2 + zc1b0 + zc1b1 + zc1b2

    def fp2_conjugate(self, a: list[ModuloCircuitElement]):
        assert len(a) == 2
        return [a[0], self.neg(a[1])]

    def mul_by_non_residue_k_power_n(
        self, x: list[ModuloCircuitElement], k: int, n: int
    ):
        from garaga.hints.tower_backup import E2

        assert len(x) == 2
        curve = CURVES[self.curve_id]
        non_residue = E2(curve.nr_a0, curve.nr_a1, curve.p)
        # nr^(n*(p^k-1)/6)
        res: E2 = non_residue ** ((n * (curve.p**k - 1) // 6))
        if res.a1 != 0:
            factor = [
                self.set_or_get_constant(res.a0),
                self.set_or_get_constant(res.a1),
            ]
            return self.fp2_mul(factor, x)
        else:
            factor = self.set_or_get_constant(res.a0)
            return [self.mul(factor, x[0]), self.mul(factor, x[1])]

    def fp12_frob(self, a: list[ModuloCircuitElement]):
        assert len(a) == 12

        c0b0 = self.fp2_conjugate(a[0:2])
        c0b0[0] = self.add(
            c0b0[0], self.set_or_get_constant(0)
        )  # Input cannot be output. # TODO: Make a manual circuit or something better.
        c0b1 = self.fp2_conjugate(a[2:4])
        c0b2 = self.fp2_conjugate(a[4:6])
        c1b0 = self.fp2_conjugate(a[6:8])
        c1b1 = self.fp2_conjugate(a[8:10])
        c1b2 = self.fp2_conjugate(a[10:12])

        c0b1 = self.mul_by_non_residue_k_power_n(c0b1, 1, 2)
        c0b2 = self.mul_by_non_residue_k_power_n(c0b2, 1, 4)
        c1b0 = self.mul_by_non_residue_k_power_n(c1b0, 1, 1)
        c1b1 = self.mul_by_non_residue_k_power_n(c1b1, 1, 3)
        c1b2 = self.mul_by_non_residue_k_power_n(c1b2, 1, 5)

        return c0b0 + c0b1 + c0b2 + c1b0 + c1b1 + c1b2

    def fp12_frob_square(self, a: list[ModuloCircuitElement]):
        assert len(a) == 12
        c0b0 = [
            self.add(a[0], self.set_or_get_constant(0)),
            self.add(a[1], self.set_or_get_constant(0)),
        ]  # Input cannot be output. # TODO: Make a manual circuit or something better.
        c0b1 = self.mul_by_non_residue_k_power_n(a[2:4], 2, 2)
        c0b2 = self.mul_by_non_residue_k_power_n(a[4:6], 2, 4)
        c1b0 = self.mul_by_non_residue_k_power_n(a[6:8], 2, 1)
        c1b1 = self.mul_by_non_residue_k_power_n(a[8:10], 2, 3)
        c1b2 = self.mul_by_non_residue_k_power_n(a[10:12], 2, 5)

        return c0b0 + c0b1 + c0b2 + c1b0 + c1b1 + c1b2

    def fp12_frob_cube(self, a: list[ModuloCircuitElement]):
        assert len(a) == 12

        c0b0 = self.fp2_conjugate(a[0:2])
        c0b0[0] = self.add(
            c0b0[0], self.set_or_get_constant(0)
        )  # Input cannot be output. # TODO: Make a manual circuit or something better.
        c0b1 = self.fp2_conjugate(a[2:4])
        c0b2 = self.fp2_conjugate(a[4:6])
        c1b0 = self.fp2_conjugate(a[6:8])
        c1b1 = self.fp2_conjugate(a[8:10])
        c1b2 = self.fp2_conjugate(a[10:12])

        c0b1 = self.mul_by_non_residue_k_power_n(c0b1, 3, 2)
        c0b2 = self.mul_by_non_residue_k_power_n(c0b2, 3, 4)
        c1b0 = self.mul_by_non_residue_k_power_n(c1b0, 3, 1)
        c1b1 = self.mul_by_non_residue_k_power_n(c1b1, 3, 3)
        c1b2 = self.mul_by_non_residue_k_power_n(c1b2, 3, 5)

        return c0b0 + c0b1 + c0b2 + c1b0 + c1b1 + c1b2

    def fp12_cyclotomic_square_compressed(
        self,
        xc0b1: list[ModuloCircuitElement],
        xc0b2: list[ModuloCircuitElement],
        xc1b0: list[ModuloCircuitElement],
        xc1b2: list[ModuloCircuitElement],
    ):
        assert len(xc0b1) == len(xc0b2) == len(xc1b0) == len(xc1b2) == 2

        t0 = self.fp2_square(xc0b1)
        t1 = self.fp2_square(xc1b2)
        t5 = self.vector_add(xc0b1, xc1b2)
        t2 = self.fp2_square(t5)

        t3 = self.vector_add(t0, t1)
        t5 = self.vector_sub(t2, t3)

        t6 = self.vector_add(xc1b0, xc0b2)
        t3 = self.fp2_square(t6)
        t2 = self.fp2_square(xc1b0)

        t6 = self.fp2_mul_by_non_residue(t5)
        t5 = self.vector_add(t6, xc1b0)
        t5 = self.vector_add(t5, t5)

        zc1b0 = self.vector_add(t5, t6)

        t4 = self.fp2_mul_by_non_residue(t1)
        t5 = self.vector_add(t0, t4)
        t6 = self.vector_sub(t5, xc0b2)

        t1 = self.fp2_square(xc0b2)
        t6 = self.vector_add(t6, t6)
        zc0b2 = self.vector_add(t6, t5)

        t4 = self.fp2_mul_by_non_residue(t1)
        t5 = self.vector_add(t2, t4)
        t6 = self.vector_sub(t5, xc0b1)
        t6 = self.vector_add(t6, t6)
        zc0b1 = self.vector_add(t6, t5)

        t0 = self.vector_add(t2, t1)
        t5 = self.vector_sub(t3, t0)
        t6 = self.vector_add(t5, xc1b2)
        t6 = self.vector_add(t6, t6)
        zc1b2 = self.vector_add(t5, t6)

        return zc0b1 + zc0b2 + zc1b0 + zc1b2

    def fp12_decompress_karabina_pt_I_c1b2_Z(self, xc0b1, xc1b2):

        assert len(xc0b1) == len(xc1b2) == 2

        t0 = self.fp2_mul(xc0b1, xc1b2)
        t0 = self.vector_add(t0, t0)

        # t1 = xc0b2

        return t0

    def fp12_decompress_karabina_pt_I_c1b2_NZ(
        self,
        xc0b1: list[ModuloCircuitElement],
        xc0b2: list[ModuloCircuitElement],
        xc1b0: list[ModuloCircuitElement],
        xc1b2: list[ModuloCircuitElement],
    ):
        assert len(xc0b1) == len(xc0b2) == len(xc1b0) == len(xc1b2) == 2

        t0 = self.fp2_square(xc0b1)
        t1 = self.vector_sub(t0, xc0b2)
        t1 = self.vector_add(t1, t1)
        t1 = self.vector_add(t1, t0)

        t2 = self.fp2_square(xc1b2)
        t0 = self.vector_add(self.fp2_mul_by_non_residue(t2), t1)
        t1 = self.vector_add(xc1b0, xc1b0)
        t1 = self.vector_add(t1, t1)

        return t0, t1

    def fp12_decompress_karabina_pt_II(
        self,
        t0: list[ModuloCircuitElement],
        t1: list[ModuloCircuitElement],
        xc0b1: list[ModuloCircuitElement],
        xc0b2: list[ModuloCircuitElement],
        xc1b0: list[ModuloCircuitElement],
        xc1b2: list[ModuloCircuitElement],
    ):
        assert len(t0) == len(t1) == 2
        assert len(xc0b1) == len(xc0b2) == len(xc1b0) == len(xc1b2) == 2

        zc1b1 = self.fp2_div(t0, t1)

        t1 = self.fp2_mul(xc0b2, xc0b1)
        t2 = self.vector_sub(self.fp2_square(zc1b1), t1)
        t2 = self.vector_add(t2, t2)
        t2 = self.vector_sub(t2, t1)
        t1 = self.fp2_mul(xc1b0, xc1b2)
        t2 = self.vector_add(t2, t1)
        zc0b0 = self.fp2_mul_by_non_residue(t2)
        zc0b0 = [
            self.add(zc0b0[0], self.set_or_get_constant(1)),
            zc0b0[1],
        ]

        return zc0b0, zc1b1



"""
 -----------------------------------------------------------------------------
 ε – Eisenstein integers in pure Python (fully commented)
 -----------------------------------------------------------------------------
 A **beginner‑friendly walk‑through**
 -----------------------------------------------------------------------------
 This file implements three things:

   • an *Eisenstein integer* class  z = a0 + a1·ω  with  ω = e^{2πi/3}
   • an *exact* Euclidean division  quo_rem()  in the ring ℤ[ω]
   • the *Half‑GCD* algorithm, one of the fastest ways to compute a greatest
     common divisor when the inputs are very large
 """

from __future__ import annotations  # ► allow the class to refer to itself in type hints

import math
from typing import Tuple

# ---------------------------------------------------------------------------
# 1.  Tiny helper for the Euclidean division in ℤ[ω]
# ---------------------------------------------------------------------------

#  The six **unit directions** of the hexagonal lattice.  Graphically:
#     ↑ (0,1)
#  ↖       ↗
# (-1,1)   (1,0)
#  ↙       ↘
# (-1,0)   (1,-1)
#     ↓ (0,-1)
_NEIGHBOURS = ((1, 0), (0, 1), (-1, 1), (-1, 0), (0, -1), (1, -1))


def _round_nearest(z: int, d: int) -> int:
    """Symmetric integer rounding  ——  ⌊ (z + d/2) / d ⌋ for *any* sign of *z*.

    • *d* is strictly positive (we only call this with norms, which are > 0).
    • For   z ≥ 0   the formula above already works.
    • For   z <  0   we cannot add  +d/2  (would move us the wrong way),
      so we flip the sign, apply the positive rule, then flip back.
    """
    half = d >> 1  # bit‑shift right by 1  ⇒  divide by 2 (fast, exact)
    return (z + half) // d if z >= 0 else -((-z + half) // d)


# ---------------------------------------------------------------------------
# 2.  The EisensteinInteger class
# ---------------------------------------------------------------------------


class EisensteinInteger:
    """Eisenstein integers:  *hexagonal* analogue of Gaussian integers.

    Definition
    ----------
    ω  (pronounced "omega") is the complex cube‑root of 1 with positive
    imaginary part:  ω = e^{2πi/3} = −½ + i·√3/2.
    It satisfies  ω² + ω + 1 = 0  and  |ω| = 1.

    Every element of  ℤ[ω]  can be written uniquely as
        z = a₀ + a₁·ω,    with  a₀, a₁ ∈ ℤ.
    The pair (a₀, a₁) are the *hex‑coordinates* in the triangular lattice.

    In code we store just those two integers.
    """

    # Class‑level constants (created *after* the class body below)
    ZERO: "EisensteinInteger"  # 0 + 0·ω
    ONE: "EisensteinInteger"  # 1 + 0·ω

    # ────────────────────────────────────────────────────────────────────
    #  Construction & basic utilities
    # ────────────────────────────────────────────────────────────────────
    def __init__(self, a0: int = 0, a1: int = 0):
        if not isinstance(a0, int) or not isinstance(a1, int):
            raise TypeError("Coefficients a0 and a1 must be integers")
        self.a0 = a0
        self.a1 = a1

    def copy(self) -> "EisensteinInteger":
        """Clone *self* (handy for algorithms that mutate a temporary copy)."""
        return EisensteinInteger(self.a0, self.a1)

    def is_zero(self) -> bool:
        return self.a0 == 0 and self.a1 == 0

    def __str__(self) -> str:
        """Pretty print   3 + 2·ω,   −ω,  0,  …"""
        if self.a1 == 0:
            return str(self.a0)

        a0_part = f"{self.a0}" if self.a0 != 0 else ""
        # coefficient of ω
        if self.a1 == 1:
            a1_part = "ω"
        elif self.a1 == -1:
            a1_part = "-ω"
        else:
            a1_part = f"{self.a1}*ω"

        if a0_part == "":
            return a1_part
        sign = " + " if self.a1 > 0 else " - "
        abs_a1_part = "ω" if abs(self.a1) == 1 else f"{abs(self.a1)}*ω"
        return f"{a0_part}{sign}{abs_a1_part}"

    def __repr__(self) -> str:
        """Unambiguous form so that  eval(repr(z)) == z  (useful in a REPL)."""
        return f"EisensteinInteger({self.a0}, {self.a1})"

    # ────────────────────────────────────────────────────────────────────
    #  Equality & negation
    # ────────────────────────────────────────────────────────────────────
    def __eq__(self, other: object) -> bool:  # called by  z1 == z2
        return (
            isinstance(other, EisensteinInteger)
            and self.a0 == other.a0
            and self.a1 == other.a1
        )

    def __neg__(self) -> "EisensteinInteger":  # −z
        return EisensteinInteger(-self.a0, -self.a1)

    # ────────────────────────────────────────────────────────────────────
    #  Conjugation
    # ────────────────────────────────────────────────────────────────────
    def conjugate(self) -> "EisensteinInteger":
        """Field automorphism  ω ↦ ω²  (complex conjugation).
        Algebraically:  conj(a₀ + a₁·ω) = (a₀ − a₁) − a₁·ω.
        """
        return EisensteinInteger(self.a0 - self.a1, -self.a1)

    # ────────────────────────────────────────────────────────────────────
    #  Addition, subtraction, multiplication
    # ────────────────────────────────────────────────────────────────────
    def __add__(self, other: "EisensteinInteger") -> "EisensteinInteger":
        return EisensteinInteger(self.a0 + other.a0, self.a1 + other.a1)

    def __sub__(self, other: "EisensteinInteger") -> "EisensteinInteger":
        return EisensteinInteger(self.a0 - other.a0, self.a1 - other.a1)

    def __mul__(self, other: object) -> "EisensteinInteger":
        """Product formula derived from  (x₀ + x₁ω)(y₀ + y₁ω).
        Expand and use  ω² = −1 − ω.
        """
        if isinstance(other, EisensteinInteger):
            x0, x1, y0, y1 = self.a0, self.a1, other.a0, other.a1
            return EisensteinInteger(
                x0 * y0 - x1 * y1,  # coefficient of 1
                x0 * y1 + x1 * y0 - x1 * y1,  # coefficient of ω
            )
        elif isinstance(other, int):
            return EisensteinInteger(self.a0 * other, self.a1 * other)
        return NotImplemented

    # Support  int * EisensteinInteger  (commutative)
    __rmul__ = __mul__

    # ────────────────────────────────────────────────────────────────────
    #  Norm (always a *plain* integer)
    # ────────────────────────────────────────────────────────────────────
    def norm(self) -> int:
        """N(a₀ + a₁·ω)  =  a₀² + a₁² − a₀·a₁   ≥ 0
        Proof: treat ω as  −½ + i·√3/2  and compute |z|².
        """
        return self.a0 * self.a0 + self.a1 * self.a1 - self.a0 * self.a1

    # -------------------------------------------------------------------
    #  Euclidean division  ——  the heart of arithmetic in ℤ[ω]
    # -------------------------------------------------------------------
    def quo_rem(
        self, y: "EisensteinInteger"
    ) -> Tuple["EisensteinInteger", "EisensteinInteger"]:
        """Return  (q, r)  such that   self = q·y + r   with  N(r) < N(y).

        Strategy (all integer‑only):
          1. Multiply by the *conjugate* of y to get an element of the
             lattice that is colinear with y – this is the usual trick to
             mimic complex division without leaving the ring.
          2. Divide the two coordinates by the *positive* norm N(y) and
             apply symmetric rounding to obtain a *first guess* q.
          3. If this guess does not yet satisfy the Euclidean inequality
             N(r) < N(y), walk at most two unit steps in the hex lattice
             until it does.

        Because ℤ[ω] is a Euclidean domain, this procedure is guaranteed
        to terminate and the while‑loop never needs more than two laps.
        """
        if y.is_zero():
            raise ZeroDivisionError("division by zero EisensteinInteger")

        nrm = y.norm()  # denominator of the coordinates ( > 0 )
        num = self * y.conjugate()  # numerator in the same lattice line

        # --- 1st approximation by independent rounding -----------------
        q0 = _round_nearest(num.a0, nrm)
        q1 = _round_nearest(num.a1, nrm)
        q = EisensteinInteger(q0, q1)
        r = self - q * y

        # --- Improve the quotient until Euclidean property holds --------
        while r.norm() >= nrm:  # very rarely true (≈ 1 in 12 cases)
            best_q, best_r, best_norm = None, None, r.norm()
            # test the 6 neighbours
            for dp, dq in _NEIGHBOURS:
                cand_q = EisensteinInteger(q0 + dp, q1 + dq)
                cand_r = self - cand_q * y
                cand_norm = cand_r.norm()
                if cand_norm < best_norm:
                    best_q, best_r, best_norm = cand_q, cand_r, cand_norm
            # move to the better neighbour (Euclidean property assures one exists)
            q, r = best_q, best_r
            q0, q1 = q.a0, q.a1  # update centre for the *next* neighbour pass
        return q, r

    # Allow the syntactic sugar  z // y   and  z % y -------------------------
    def __floordiv__(self, other: "EisensteinInteger") -> "EisensteinInteger":
        return self.quo_rem(other)[0]

    def __mod__(self, other: "EisensteinInteger") -> "EisensteinInteger":
        return self.quo_rem(other)[1]


# -------- class constants (must be assigned *after* the class is complete) ---
EisensteinInteger.ZERO = EisensteinInteger(0, 0)
EisensteinInteger.ONE = EisensteinInteger(1, 0)

# ---------------------------------------------------------------------------
# 3.  Half‑GCD  ——  sub‑quadratic gcd via size halving
# ---------------------------------------------------------------------------


def half_gcd(
    a: EisensteinInteger,
    b: EisensteinInteger,
) -> Tuple[EisensteinInteger, EisensteinInteger, EisensteinInteger]:
    """Return  (w, v, u)  with  w = a·u + b·v  and  N(w) < √N(a).

    Why is that useful?  The classical Euclidean gcd repeatedly replaces the
    *larger* input by its remainder modulo the smaller one – this shrinks one
    argument at every step, but only by a *constant* factor, so it needs
    O(log N) iterations.

    **Half‑GCD** knocks out *half* the size (in bits) at every major step and
    therefore runs in roughly O(log N / log log N) time.

    The idea is to run the Euclidean algorithm *just long enough* for the
    remainder’s norm to dip below √N(a).  The output coefficients u, v let us
    continue assembling the gcd later (the full gcd algorithm calls half_gcd
    recursively on smaller and smaller inputs).
    """
    if not isinstance(a, EisensteinInteger) or not isinstance(b, EisensteinInteger):
        raise TypeError("Inputs must be EisensteinInteger instances")

    # Extended Euclidean bookkeeping matrices
    a_run, b_run = a.copy(), b.copy()
    u, v = EisensteinInteger.ONE.copy(), EisensteinInteger.ZERO.copy()
    u_, v_ = EisensteinInteger.ZERO.copy(), EisensteinInteger.ONE.copy()
    # Invariants throughout the loop:
    #     a_run =  a·u  + b·v
    #     b_run =  a·u_ + b·v_

    limit = math.isqrt(a.norm())  # √N(a)  (integer square root)

    # Safety valve: more than ~2e5 iterations would imply a bug
    for _ in range(200_000):
        if b_run.norm() < limit or b_run.is_zero():
            break
        q, r = a_run.quo_rem(b_run)

        # update (a_run, b_run)  and the coefficient matrix simultaneously
        a_run, b_run = b_run, r
        u, u_ = u_, u - q * u_
        v, v_ = v_, v - q * v_
    else:
        raise RuntimeError("Half‑GCD failed to converge – too many steps")

    # *Return order* follows the classic Half‑GCD papers:  (w, v, u)
    return b_run, v_, u_


if __name__ == "__main__":
    z1 = EisensteinInteger(3, 2)  # 3 + 2ω
    z2 = EisensteinInteger(1, -1)  # 1 - ω

    print(f"z1 = {z1}")
    print(f"z2 = {z2}")
    print(f"z1 + z2 = {z1 + z2}")
    print(f"z1 - z2 = {z1 - z2}")
    print(f"z1 * z2 = {z1 * z2}")
    print(f"Norm(z1) = {z1.norm()}")
    print(f"Conj(z1) = {z1.conjugate()}")
    print(f"-z1 = {-z1}")

    q, r = z1.quo_rem(z2)
    print(f"z1 // z2 = {q}")
    print(f"z1 % z2 = {r}")
    print(f"Check: z2 * q + r = {z2 * q + r}, Expected: {z1}")

    # Half GCD example
    a = EisensteinInteger(10, 3)
    b = EisensteinInteger(3, 1)
    print(f"\nHalf GCD for a = {a}, b = {b}")
    w, v_res, u_res = half_gcd(a, b)
    print(f"  w = {w}")
    print(f"  v = {v_res}")
    print(f"  u = {u_res}")
    check = a * u_res + b * v_res
    print(f"  Check: a*u + b*v = {check}, Expected w: {w}")
    print(f"  Norm(w) = {w.norm()}, Limit = {math.isqrt(a.norm())}")

    # a = EisensteinInteger(8+7j * (math.sqrt(3)/2 + 0.5j)) # Approximate large number
    a_int = EisensteinInteger(13, 14)  # Example large int coeffs
    b_int = EisensteinInteger(5, 3)
    print(f"\nHalf GCD for a = {a_int}, b = {b_int}")
    w, v_res, u_res = half_gcd(a_int, b_int)
    print(f"  w = {w}")
    print(f"  v = {v_res}")
    print(f"  u = {u_res}")
    check = a_int * u_res + b_int * v_res
    print(f"  Check: a*u + b*v = {check}, Expected w: {w}")
    print(f"  Norm(w) = {w.norm()}, Limit = {math.isqrt(a_int.norm())}")



import garaga.hints.io as io
from garaga import garaga_rs
from garaga.drand.client import (
    DrandNetwork,
    G2Point,
    digest_func,
    get_chain_info,
    get_randomness,
)
from garaga.signature import hash_to_curve
from garaga.starknet.tests_and_calldata_generators.map_to_curve import *
from garaga.starknet.tests_and_calldata_generators.mpcheck import (
    G1G2Pair,
    MPCheckCalldataBuilder,
)


def drand_round_to_calldata(round_number: int, use_rust=False) -> list[int]:
    if use_rust:
        return _drand_round_to_calldata_rust(round_number)

    message = digest_func(round_number)
    # print(f"round {round_number} message {message}")
    msg_point = hash_to_curve(message, CurveID.BLS12_381, "sha256")

    chain = get_chain_info(DrandNetwork.quicknet.value)

    round = get_randomness(chain.hash, round_number)

    sig_pt = round.signature_point
    ###################
    mpc_builder = MPCheckCalldataBuilder(
        curve_id=CurveID.BLS12_381,
        pairs=[
            G1G2Pair(p=sig_pt, q=G2Point.get_nG(CurveID.BLS12_381, 1)),
            G1G2Pair(p=msg_point, q=-chain.public_key),
        ],
        n_fixed_g2=2,
        public_pair=None,
    )

    cd = []
    sig_pt: G1Point
    cd.append(round_number)
    cd.extend(io.bigint_split(sig_pt.x))
    cd.extend(io.bigint_split(sig_pt.y))
    cd.extend(build_hash_to_curve_hint(message).to_calldata())
    cd.extend(mpc_builder.serialize_to_calldata())

    return [len(cd)] + cd


def _drand_round_to_calldata_rust(
    round_number: int,
) -> list[int]:
    chain = get_chain_info(DrandNetwork.quicknet.value)
    round = get_randomness(chain.hash, round_number)
    randomness = round.randomness
    signature = int(round.signature, 16)
    data = [
        round_number,
        randomness,
        signature,
    ]
    return garaga_rs.drand_calldata_builder(data)


if __name__ == "__main__":

    drand_round_to_calldata(1)
    drand_round_to_calldata(2)
    # drand_round_to_calldata(3)



from garaga.algebra import Fp2, ModuloCircuitElement
from garaga.hints.extf_mul import nondeterministic_extension_field_div
from garaga.modulo_circuit import ModuloCircuit, WriteOps


class Fp2Circuits(ModuloCircuit):
    """
    Implements circuits for Fp2 field operations.
    """

    def __init__(
        self,
        name: str,
        curve_id: int,
        generic_circuit: bool = False,
        compilation_mode: int = 0,
    ):
        super().__init__(
            name=name,
            curve_id=curve_id,
            generic_circuit=generic_circuit,
            compilation_mode=compilation_mode,
        )

    def fp2_is_non_zero(
        self, a: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        """
        Returns [1,0] if a ≠ 0, [0,0] if a == 0, working in Fp2.
        An Fp2 element is non-zero if either its real or imaginary part is non-zero.
        """
        # Check if real part is non-zero
        real_is_non_zero = self.fp_is_non_zero(a[0])

        # Check if imaginary part is non-zero
        imag_is_non_zero = self.fp_is_non_zero(a[1])

        # Either part must be non-zero for the Fp2 element to be non-zero
        # Using 1-(1-a)(1-b) = a + b - ab to compute OR
        result = self.sub(
            self.add(real_is_non_zero, imag_is_non_zero),
            self.mul(real_is_non_zero, imag_is_non_zero),
        )

        # Return as Fp2 element [result, 0]
        return [result, self.set_or_get_constant(0)]

    def fp2_mul(self, X: list[ModuloCircuitElement], Y: list[ModuloCircuitElement]):
        # Assumes the irreducible poly is X^2 + 1.
        assert len(X) == len(Y) == 2 and all(
            isinstance(x, ModuloCircuitElement) and isinstance(y, ModuloCircuitElement)
            for x, y in zip(X, Y)
        )
        # xy = (x0 + i*x1) * (y0 + i*y1) = (x0*y0 - x1*y1) + i * (x0*y1 + x1*y0)
        return [
            self.sub(
                self.mul(X[0], Y[0], comment="Fp2 mul start"),
                self.mul(X[1], Y[1]),
                comment="Fp2 mul real part end",
            ),
            self.add(
                self.mul(X[0], Y[1]),
                self.mul(X[1], Y[0]),
                comment="Fp2 mul imag part end",
            ),
        ]

    def fp2_mul_by_non_residue(self, X: list[ModuloCircuitElement]):
        assert len(X) == 2 and all(isinstance(x, ModuloCircuitElement) for x in X)
        if self.curve_id == 1:
            # Non residue (1,1)
            # (a0 + i*a1) * (1 + i)
            a_tmp = self.add(X[0], X[1])
            a = self.add(a_tmp, a_tmp)
            b = X[0]
            z_a0 = self.sub(b, X[1])
            z_a1 = self.sub(self.sub(a, b), X[1])
            return [z_a0, z_a1]
        elif self.curve_id == 0:
            # Non residue (9, 1)
            # (a0 + i*a1) * (9 + i)
            a_tmp = self.add(X[0], X[1])
            a = self.mul(a_tmp, self.set_or_get_constant(10))
            b = self.mul(X[0], self.set_or_get_constant(9))
            z_a0 = self.sub(b, X[1])
            z_a1 = self.sub(self.sub(a, b), X[1])
            return [z_a0, z_a1]

        else:
            raise ValueError(
                f"Unsupported curve id for fp2 mul by non residue: {self.curve_id}"
            )

    def fp2_square(self, X: list[ModuloCircuitElement]):
        # Assumes the irreducible poly is X^2 + 1.
        # x² = (x0 + i x1)² = (x0² - x1²) + 2 * i * x0 * x1 = (x0+x1)(x0-x1) + i * 2 * x0 * x1.
        # (x0+x1)*(x0-x1) is cheaper than x0² - x1². (2 ADD + 1 MUL) vs (1 ADD + 2 MUL) (16 vs 20 steps)
        assert len(X) == 2 and all(isinstance(x, ModuloCircuitElement) for x in X)
        return [
            self.mul(self.add(X[0], X[1]), self.sub(X[0], X[1])),
            self.double(self.mul(X[0], X[1])),
        ]

    def fp2_sqrt(
        self, element: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        """
        Computes "a" square root of a field element.
        /!\ Warning : This circuit is non deterministic /!\
        /!\ Two square roots exist for any non-zero element, and no constraint is enforced to select any of them /!\
        Raises ValueError if the element is not a quadratic residue.
        """
        assert self.compilation_mode == 0, "fp2_sqrt is not supported in cairo 1 mode"

        root = Fp2(element[0].felt, element[1].felt).sqrt()
        if root is None:
            raise ValueError("No square root found for the given element")

        if not root.lexicographically_largest:
            root = Fp2.zero(element[0].p).__sub__(root)

        root = self.write_elements([root.a0, root.a1], WriteOps.WITNESS)
        self.fp2_mul_and_assert(root, root, element, comment="Fp2 sqrt")
        return root

    def fp2_inv(self, X: list[ModuloCircuitElement]):
        assert len(X) == 2 and all(isinstance(x, ModuloCircuitElement) for x in X)
        t0 = self.mul(X[0], X[0], comment="Fp2 Inv start")
        t1 = self.mul(X[1], X[1])
        t0 = self.add(t0, t1)
        t1 = self.inv(t0)
        inv0 = self.mul(X[0], t1, comment="Fp2 Inv real part end")
        inv1 = self.neg(self.mul(X[1], t1), comment="Fp2 Inv imag part end")
        return [inv0, inv1]

    def fp2_div(self, X: list[ModuloCircuitElement], Y: list[ModuloCircuitElement]):
        assert len(X) == len(Y) == 2 and all(
            isinstance(x, ModuloCircuitElement) and isinstance(y, ModuloCircuitElement)
            for x, y in zip(X, Y)
        )
        if self.compilation_mode == 0:
            x_over_y = nondeterministic_extension_field_div(X, Y, self.curve_id, 2)
            x_over_y = self.write_elements(x_over_y, WriteOps.WITNESS)
            # x_over_y = d0 + i * d1
            # y = y0 + i * y1
            # x = x_over_y*y = d0*y0 - d1*y1 + i * (d0*y1 + d1*y0)
            self.sub_and_assert(
                a=self.mul(x_over_y[0], Y[0]), b=self.mul(x_over_y[1], Y[1]), c=X[0]
            )
            self.add_and_assert(
                a=self.mul(x_over_y[0], Y[1]), b=self.mul(x_over_y[1], Y[0]), c=X[1]
            )
            return x_over_y
        elif self.compilation_mode == 1:
            # Todo : consider passing as calldata if possible.
            inv = self.fp2_inv(Y)
            return self.fp2_mul(X, inv)

    def fp2_frobenius_map(
        self, element: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        """
        Applies the Frobenius map to an element in F_{p^2}.
        For an element a + bi, returns a - bi.

        This is because:
        (a + bi)^p = a^p + (bi)^p = a^p + b^p * i^p
        In Fp: a^p = a, b^p = b
        In Fp2: i^p = -i (p is odd > 2)
        Therefore: (a + bi)^p = a - bi
        """
        a, b = element  # element = a + bi
        return [a, self.neg(b)]  # return a - bi

    def fp2_add(
        self, X: list[ModuloCircuitElement], Y: list[ModuloCircuitElement]
    ) -> list[ModuloCircuitElement]:
        # Assumes elements are represented as pairs (a + bi)
        return self.vector_add(X, Y)

    def fp2_sub(self, X: list[ModuloCircuitElement], Y: list[ModuloCircuitElement]):
        # Assumes elements are represented as pairs (a + bi)
        return self.vector_sub(X, Y)

    def fp2_mul_and_assert(
        self,
        a: list[ModuloCircuitElement],
        b: list[ModuloCircuitElement],
        c: list[ModuloCircuitElement],
        comment: str | None = None,
    ):
        """
        Multiplies two Fp2 elements a and b and asserts the result equals c.
        For a = (a0 + i*a1) and b = (b0 + i*b1), asserts:
        c0 = a0*b0 - a1*b1
        c1 = a0*b1 + a1*b0
        """
        assert (
            self.compilation_mode == 0
        ), "fp2_mul_and_assert is not supported in cairo 1 mode"

        assert len(a) == len(b) == len(c) == 2, "Fp2 elements must be length 2"

        # Calculate intermediate products
        a0b0 = self.mul(a[0], b[0], comment=f"{comment}: a0*b0" if comment else None)
        a1b1 = self.mul(a[1], b[1], comment=f"{comment}: a1*b1" if comment else None)
        a0b1 = self.mul(a[0], b[1], comment=f"{comment}: a0*b1" if comment else None)
        a1b0 = self.mul(a[1], b[0], comment=f"{comment}: a1*b0" if comment else None)

        # Assert real part: c0 = a0*b0 - a1*b1
        self.add_and_assert(
            a=a0b0,
            b=self.neg(a1b1),
            c=c[0],
            comment=f"{comment}: assert real part" if comment else None,
        )

        # Assert imaginary part: c1 = a0*b1 + a1*b0
        self.add_and_assert(
            a=a0b1,
            b=a1b0,
            c=c[1],
            comment=f"{comment}: assert imaginary part" if comment else None,
        )

        return c

    def fp2_eval_horner(
        self,
        poly: list[ModuloCircuitElement],  # [a0_real, a0_imag, a1_real, a1_imag, ...]
        z: list[ModuloCircuitElement],  # z = [real, imag]
        poly_name: str = None,
        var_name: str = "z",
    ) -> list[ModuloCircuitElement]:
        """
        Evaluates a polynomial with Fp2 coefficients at point z using Horner's method.
        Assumes that the polynomial is in the form a0 + a1*z + a2*z^2 + ... + an*z^n,
        indexed with the constant coefficient first.
        Coefficients are stored in a flat array where each consecutive pair represents
        the real and imaginary parts of an Fp2 element.

        Args:
            poly: Flat list of coefficients [a0_real, a0_imag, a1_real, a1_imag, ...]
            z: The Fp2 point to evaluate at, represented as [real, imag]
            poly_name: Optional name for debugging
            var_name: Optional variable name for debugging

        Returns:
            [real, imag] representing the result in Fp2
        """
        if poly_name is None:
            poly_name = "UnnamedPoly"

        assert len(poly) % 2 == 0, "Polynomial coefficients array must have even length"
        n_coeffs = len(poly) // 2

        # Start with the highest degree coefficient
        acc = [poly[-2], poly[-1]]  # Get last pair [real, imag]

        # Iterate through remaining coefficients in reverse order
        for i in range(n_coeffs - 2, -1, -1):
            acc = self.fp2_add(
                self.fp2_mul(
                    acc,
                    z,
                ),
                [poly[2 * i], poly[2 * i + 1]],  # Get i-th coefficient pair
            )

        return acc
